# Copyright 2019-2023 The TFP CausalImpact Authors
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Import commonly used modules and classes so user doesn't have to."""

__version__ = "0.2.0"

import os

# TF 2.12 currently generates a huge number of useless log.info messages
# directly from C++, which cannot be filtered using standard Python logging
# tools -- see https://github.com/tensorflow/tensorflow/issues/59779 .
# To improve the TFP CausalImpact user experience, we suppress these messages.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

# pylint: disable=g-import-not-at-top
from causalimpact.causalimpact_lib import CausalImpactAnalysis
from causalimpact.causalimpact_lib import DataOptions
from causalimpact.causalimpact_lib import fit_causalimpact
from causalimpact.causalimpact_lib import InferenceOptions
from causalimpact.causalimpact_lib import ModelOptions
from causalimpact.causalimpact_lib import Seasons
from causalimpact.indices import InputDateType
from causalimpact.plot import plot
from causalimpact.summary import summary



# Copyright 2020-2023 The TFP CausalImpact Authors
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for indices.py."""

import os
from typing import Tuple

from absl.testing import absltest
from absl.testing import parameterized

from causalimpact import indices

import numpy as np
import pandas as pd


CURR_PATH = os.path.dirname(__file__)
TEST_DATA = os.path.join(CURR_PATH, "testdata", "data.csv")


class WrapperTest(parameterized.TestCase):
  """Tests for indices."""

  @classmethod
  def setUpClass(cls):
    super(WrapperTest, cls).setUpClass()

    # Read in sample data and set column "t" as the index
    with open(TEST_DATA, "r") as fl:
      df = pd.read_csv(fl)
    df = df.set_index(pd.to_datetime(df["t"]))
    df.drop(columns=["t"], inplace=True)

    cls.data = df

  @parameterized.named_parameters([
      {
          "testcase_name": "dates_before_the_data",
          "period": (pd.to_datetime("2016-01-20 22:41:20"),
                     pd.to_datetime("2016-01-21 22:41:20")),
          "error_string": "Aligned period end not found in the index."
      },
      {
          "testcase_name": "dates_after_the_data",
          "period": (pd.to_datetime("2022-01-20 22:41:20"),
                     pd.to_datetime("2022-01-21 22:41:20")),
          "error_string": "Aligned period start not found in the index."
      },
      {
          "testcase_name": "dates_not_in_order",
          "period": (pd.to_datetime("2016-02-20 22:41:50"),
                     pd.to_datetime("2016-02-20 22:41:20")),
          "error_string": "Period end must be after period start. "
      },
  ])
  def testCheckPeriods(self, period: Tuple[str], error_string: str):
    with self.assertRaisesRegex(ValueError, error_string):
      indices._check_period(period, self.data)

  @parameterized.named_parameters([
      {
          "testcase_name": "overlapping_periods",
          "pre_period": (pd.to_datetime("2016-02-20 22:41:20"),
                         pd.to_datetime("2016-02-20 22:41:50")),
          "post_period": (pd.to_datetime("2016-02-20 22:41:40"),
                          pd.to_datetime("2016-02-20 22:41:50")),
          "error_string": "pre_period and post_period cannot overlap."
      },
      {
          "testcase_name": "pre_period_too_short",
          "pre_period": (pd.to_datetime("2016-02-20 22:41:20"),
                         pd.to_datetime("2016-02-20 22:41:30")),
          "post_period": (pd.to_datetime("2016-02-20 22:41:40"),
                          pd.to_datetime("2016-02-20 22:41:50")),
          "error_string": "pre_period must span at least 3 time points."
      },
  ])
  def testParseAndValidatePeriods(self, pre_period: Tuple[str],
                                  post_period: Tuple[str], error_string: str):
    with self.assertRaisesRegex(ValueError, error_string):
      indices._parse_and_validate_periods(  # pylint: disable=protected-access
          pre_period, post_period, self.data)

  @parameterized.named_parameters([
      {
          "testcase_name": "_integer_periods",
          "pre_period": (0, 10),
          "post_period": (11, 90)
      },
      {
          "testcase_name": "_string_periods",
          "pre_period": ("2016-02-20 22:41:20", "2016-02-20 22:43:00"),
          "post_period": ("2016-02-20 22:43:10", "2016-02-20 22:56:20"),
      },
      {
          "testcase_name": "_string_periods_with_rounding",
          "pre_period": ("2016-02-20 22:41:15", "2016-02-20 22:43:05"),
          "post_period": ("2016-02-20 22:43:06", "2016-02-20 22:56:28"),
      },
      {
          "testcase_name":
              "_datetime_periods",
          "pre_period": (pd.to_datetime("2016-02-20 22:41:20"),
                         pd.to_datetime("2016-02-20 22:43:00")),
          "post_period": (pd.to_datetime("2016-02-20 22:43:10"),
                          pd.to_datetime("2016-02-20 22:56:20")),
      },
      {
          "testcase_name":
              "_datetime_periods_with_rounding",
          "pre_period": (pd.to_datetime("2016-02-20 22:41:11"),
                         pd.to_datetime("2016-02-20 22:43:07")),
          "post_period": (pd.to_datetime("2016-02-20 22:43:09"),
                          pd.to_datetime("2016-02-20 22:56:23")),
      },

  ])
  def testParseAndValidateData(self, pre_period, post_period):
    """Test different datetime arguments are correctly handled."""
    pre_period, post_period = indices.parse_and_validate_date_data(
        data=self.data, pre_intervention_period=pre_period, post_intervention_period=post_period)

    self.assertTupleEqual(pre_period, (pd.to_datetime("2016-02-20 22:41:20"),
                                       pd.to_datetime("2016-02-20 22:43:00")))
    self.assertTupleEqual(post_period, (pd.to_datetime("2016-02-20 22:43:10"),
                                        pd.to_datetime("2016-02-20 22:56:20")))

  @parameterized.named_parameters([{
      "testcase_name": "_integer_periods",
      "pre_period": (0, 10),
      "post_period": (11, 90)
  }])
  def testParseAndValidateNonDatetimeData(self, pre_period, post_period):
    """Test in cases data has an integer index."""
    data = self.data.copy()
    data.index = np.arange(data.shape[0])
    pre_period, post_period = indices.parse_and_validate_date_data(
        data=data, pre_intervention_period=pre_period, post_intervention_period=post_period)

    self.assertTupleEqual(pre_period, (0, 10))
    self.assertTupleEqual(post_period, (11, 90))


if __name__ == "__main__":
  absltest.main()



# Copyright 2020-2023 The TFP CausalImpact Authors
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for plot.py."""

from absl.testing import absltest
from absl.testing import parameterized

import causalimpact as ci
from causalimpact.plot import _create_plot_component_df
import numpy as np
import pandas as pd


expected_classic_dict_four_vlines = {
    "facet": {
        "row": {
            "type": "nominal",
            "field": "scale_pretty",
            "sort": ["Original", "Pointwise", "Cumulative"],
            "title": ""
        }
    },
    "spec": {
        "layer": [{
            "mark": "line",
            "encoding": {
                "color": {
                    "type": "nominal",
                    "field": "stat_pretty",
                    "legend": {
                        "labelFontSize": 16,
                        "symbolSize": 160,
                        "title": ""
                    }
                },
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "value",
                    "scale": {
                        "zero": False
                    },
                    "title": ""
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": {
                "type": "area",
                "opacity": 0.3
            },
            "encoding": {
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "upper"
                },
                "y2": {
                    "field": "lower"
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": "rule",
            "encoding": {
                "y": {
                    "type": "quantitative",
                    "field": "zero"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "pre_period_start"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "pre_period_end"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "post_period_start"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "post_period_end"
                }
            }
        }]
    },
    "resolve": {
        "scale": {
            "y": "independent"
        }
    }
}

expected_classic_dict_two_vlines = {
    "facet": {
        "row": {
            "type": "nominal",
            "field": "scale_pretty",
            "sort": ["Original", "Pointwise", "Cumulative"],
            "title": ""
        }
    },
    "spec": {
        "layer": [{
            "mark": "line",
            "encoding": {
                "color": {
                    "type": "nominal",
                    "field": "stat_pretty",
                    "legend": {
                        "labelFontSize": 16,
                        "symbolSize": 160,
                        "title": ""
                    }
                },
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "value",
                    "scale": {
                        "zero": False
                    },
                    "title": ""
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": {
                "type": "area",
                "opacity": 0.3
            },
            "encoding": {
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "upper"
                },
                "y2": {
                    "field": "lower"
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": "rule",
            "encoding": {
                "y": {
                    "type": "quantitative",
                    "field": "zero"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "pre_period_end"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "post_period_start"
                }
            }
        }]
    },
    "resolve": {
        "scale": {
            "y": "independent"
        }
    }
}

expected_classic_dict_one_vline = {
    "facet": {
        "row": {
            "type": "nominal",
            "field": "scale_pretty",
            "sort": ["Original", "Pointwise", "Cumulative"],
            "title": ""
        }
    },
    "spec": {
        "layer": [{
            "mark": "line",
            "encoding": {
                "color": {
                    "type": "nominal",
                    "field": "stat_pretty",
                    "legend": {
                        "labelFontSize": 16,
                        "symbolSize": 160,
                        "title": ""
                    }
                },
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "value",
                    "scale": {
                        "zero": False
                    },
                    "title": ""
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": {
                "type": "area",
                "opacity": 0.3
            },
            "encoding": {
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "upper"
                },
                "y2": {
                    "field": "lower"
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": "rule",
            "encoding": {
                "y": {
                    "type": "quantitative",
                    "field": "zero"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "post_period_start"
                }
            }
        }]
    },
    "resolve": {
        "scale": {
            "y": "independent"
        }
    }
}

expected_classic_dict_one_vline_integer_index = {
    "facet": {
        "row": {
            "type": "nominal",
            "field": "scale_pretty",
            "sort": ["Original", "Pointwise", "Cumulative"],
            "title": ""
        }
    },
    "spec": {
        "layer": [{
            "mark": "line",
            "encoding": {
                "color": {
                    "type": "nominal",
                    "field": "stat_pretty",
                    "legend": {
                        "labelFontSize": 16,
                        "symbolSize": 160,
                        "title": ""
                    }
                },
                "x": {
                    "type": "quantitative",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "value",
                    "scale": {
                        "zero": False
                    },
                    "title": ""
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": {
                "type": "area",
                "opacity": 0.3
            },
            "encoding": {
                "x": {
                    "type": "quantitative",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "upper"
                },
                "y2": {
                    "field": "lower"
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": "rule",
            "encoding": {
                "y": {
                    "type": "quantitative",
                    "field": "zero"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "quantitative",
                    "field": "post_period_start"
                }
            }
        }]
    },
    "resolve": {
        "scale": {
            "y": "independent"
        }
    }
}

expected_top_dict = {
    "facet": {
        "row": {
            "type": "nominal",
            "field": "scale_pretty",
            "sort": ["Original", "Pointwise", "Cumulative"],
            "title": ""
        }
    },
    "spec": {
        "layer": [{
            "mark": "line",
            "encoding": {
                "color": {
                    "type": "nominal",
                    "field": "stat_pretty",
                    "legend": {
                        "labelFontSize": 16,
                        "symbolSize": 160,
                        "title": ""
                    }
                },
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "value",
                    "scale": {
                        "zero": False
                    },
                    "title": ""
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": {
                "type": "area",
                "opacity": 0.3
            },
            "encoding": {
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "upper"
                },
                "y2": {
                    "field": "lower"
                }
            },
            "height": 200,
            "selection": {
                "selector001": {
                    "type": "interval",
                    "encodings": ["x"]
                }
            },
            "width": 600
        }, {
            "mark": "rule",
            "encoding": {
                "y": {
                    "type": "quantitative",
                    "field": "zero"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "pre_period_end"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "post_period_start"
                }
            }
        }]
    },
    "resolve": {
        "scale": {
            "y": "independent"
        }
    }
}

expected_bot_dict = {
    "facet": {
        "row": {
            "type": "nominal",
            "field": "scale_pretty",
            "sort": ["Original", "Pointwise", "Cumulative"],
            "title": ""
        }
    },
    "spec": {
        "layer": [{
            "mark": "line",
            "encoding": {
                "color": {
                    "condition": {
                        "type": "nominal",
                        "field": "stat_pretty",
                        "legend": None,
                        "selection": "selector002"
                    },
                    "value": "lightgray"
                },
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "scale": {
                        "domain": {
                            "selection": "selector001"
                        }
                    },
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "value",
                    "scale": {
                        "zero": False
                    },
                    "title": ""
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": {
                "type": "area",
                "opacity": 0.3
            },
            "encoding": {
                "x": {
                    "type": "temporal",
                    "field": "time",
                    "scale": {
                        "domain": {
                            "selection": "selector001"
                        }
                    },
                    "title": "Time"
                },
                "y": {
                    "type": "quantitative",
                    "field": "upper"
                },
                "y2": {
                    "field": "lower"
                }
            },
            "height": 200,
            "width": 600
        }, {
            "mark": "rule",
            "encoding": {
                "y": {
                    "type": "quantitative",
                    "field": "zero"
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "pre_period_end",
                    "scale": {
                        "domain": {
                            "selection": "selector001"
                        }
                    }
                }
            }
        }, {
            "mark": {
                "type": "rule",
                "strokeDash": [5, 5]
            },
            "encoding": {
                "color": {
                    "value": "grey"
                },
                "x": {
                    "type": "temporal",
                    "field": "post_period_start",
                    "scale": {
                        "domain": {
                            "selection": "selector001"
                        }
                    }
                }
            }
        }]
    },
    "resolve": {
        "scale": {
            "y": "independent"
        }
    }
}

expected_legend_dict = {
    "mark": "point",
    "encoding": {
        "color": {
            "condition": {
                "type": "nominal",
                "field": "stat_pretty",
                "legend": None,
                "selection": "selector002"
            },
            "value": "lightgray"
        },
        "y": {
            "type": "nominal",
            "axis": {
                "orient": "right"
            },
            "field": "stat_pretty",
            "title": ""
        }
    },
    "selection": {
        "selector002": {
            "type": "multi",
            "fields": ["stat_pretty"]
        }
    }
}


class PlotTest(parameterized.TestCase):

  @classmethod
  def setUpClass(cls):
    super(PlotTest, cls).setUpClass()

    # Create a fake impact time series dataframe.
    n_time_points = 10
    time_index = pd.date_range("2018-01-01", periods=n_time_points, freq="D")
    observed = pd.Series([0, 1, 2, 3, 4, 8, 9, 10, 11, 12])
    posterior_mean = pd.Series(list(range(n_time_points)))
    posterior_median = pd.Series(list(x + 0.1 for x in range(n_time_points)))
    posterior_lower = pd.Series(list(x - 0.2 for x in range(n_time_points)))
    posterior_upper = pd.Series(list(x + 0.2 for x in range(n_time_points)))
    posterior_std = pd.Series(np.repeat(0.1, repeats=n_time_points))
    point_effects_mean = pd.Series([0, 0, 0, 0, 0, 3, 3, 3, 3, 3])
    point_effects_lower = pd.Series([0, 0, 0, 0, 0, 2.8, 2.8, 2.8, 2.8, 2.8])
    point_effects_upper = pd.Series([0, 0, 0, 0, 0, 3.2, 3.2, 3.2, 3.2, 3.2])
    point_effects_std = pd.Series(np.repeat(0.1, repeats=n_time_points))
    cumulative_effects_mean = pd.Series([0, 0, 0, 0, 0, 3, 6, 9, 12, 15])
    cumulative_effects_lower = pd.Series(
        [0, 0, 0, 0, 0, 2.8, 5.6, 8.4, 11.2, 14])
    cumulative_effects_upper = pd.Series(
        [0, 0, 0, 0, 0, 3.2, 6.4, 9.6, 12.8, 16])
    cumulative_effects_std = pd.Series(np.repeat(0.1, repeats=n_time_points))
    df = pd.concat([
        observed, posterior_mean, posterior_median, posterior_lower,
        posterior_upper, posterior_std, point_effects_mean, point_effects_lower,
        point_effects_upper, point_effects_std, cumulative_effects_mean,
        cumulative_effects_lower, cumulative_effects_upper,
        cumulative_effects_std
    ],
                   axis=1)
    df.columns = [
        "observed", "posterior_mean", "posterior_median", "posterior_lower",
        "posterior_upper", "posterior_std", "point_effects_mean",
        "point_effects_lower", "point_effects_upper", "point_effects_std",
        "cumulative_effects_mean", "cumulative_effects_lower",
        "cumulative_effects_upper", "cumulative_effects_std"
    ]
    df.index = time_index
    series_1 = df.copy()
    series_1["pre_period_start"] = series_1.index[0]
    series_1["pre_period_end"] = series_1.index[3]
    series_1["post_period_start"] = series_1.index[4]
    series_1["post_period_end"] = series_1.index[-1]
    cls.ci_data_1 = ci.CausalImpactAnalysis(
        series=series_1, summary=pd.DataFrame(), posterior_samples=[])
    series_2 = df.copy()
    series_2["pre_period_start"] = series_2.index[0]
    series_2["pre_period_end"] = series_2.index[3]
    series_2["post_period_start"] = series_2.index[6]
    series_2["post_period_end"] = series_2.index[-1]
    cls.ci_data_2 = ci.CausalImpactAnalysis(
        series=series_2, summary=pd.DataFrame(), posterior_samples=[])

    # This is constructed to have start and end of pre-period and post-period
    # drawn by having a point before/intbetween/after each of them.
    series_4 = df.copy()
    series_4["pre_period_start"] = series_4.index[1]
    series_4["pre_period_end"] = series_4.index[3]
    series_4["post_period_start"] = series_4.index[6]
    series_4["post_period_end"] = series_4.index[-2]
    cls.ci_data_4 = ci.CausalImpactAnalysis(
        series=series_4, summary=pd.DataFrame(), posterior_samples=[])

    series_integer_index = df.copy()
    series_integer_index.index = pd.RangeIndex(stop=n_time_points)
    series_integer_index["pre_period_start"] = series_integer_index.index[0]
    series_integer_index["pre_period_end"] = series_integer_index.index[3]
    series_integer_index["post_period_start"] = series_integer_index.index[4]
    series_integer_index["post_period_end"] = series_integer_index.index[-1]
    cls.ci_data_integer_index = ci.CausalImpactAnalysis(
        series=series_integer_index,
        summary=pd.DataFrame(),
        posterior_samples=[])

  def testCreatePlotComponentDF_lines(self):
    lines_df = _create_plot_component_df(
        self.ci_data_1.series, component="lines")
    expected_cols = [
        "time", "post_period_start", "post_period_end", "pre_period_start",
        "pre_period_end", "value", "scale", "stat"
    ]
    self.assertSetEqual(set(expected_cols), set(lines_df.columns))

  @parameterized.named_parameters([
      {
          "testcase_name": "quantile",
          "component": "bands",
          "method": "quantile"
      },
      {
          "testcase_name": "std",
          "component": "std",
          "method": "std"
      },
  ])
  def testCreatePlotComponentDF_bands(self, component, method):
    bands_df = _create_plot_component_df(
        self.ci_data_2.series, component=component)
    expected_cols = [
        "time", "post_period_start", "post_period_end", "pre_period_start",
        "pre_period_end", "lower", "upper", "scale", "band_method"
    ]
    self.assertSetEqual(set(expected_cols), set(bands_df.columns))

    # `bands_df` has a column specifying which method ("quantile" or "std") was
    # used to create the uncertainty intervals; here we make sure the correct
    # method is given.
    self.assertEqual(bands_df["band_method"].unique(), method)

  def testPlotMatplotlib(self):
    # Create plot object and use to_dict() to convert the plot components into
    # a more easily queried dict.
    fig = ci.plot(self.ci_data_1, backend="matplotlib")
    self.assertIsNotNone(fig)

  def testClassicPlot_one_vline(self):
    # Create plot object and use to_dict() to convert the plot components into
    # a more easily queried dict.
    classic_plot_dict = ci.plot(self.ci_data_1).to_dict()

    # Check the important elements: that facets are mapped to the correct
    # variable, the plot layers (lines, bands, etc) have the correct specs, and
    # the scale resolutions are the same.
    keys_to_test = ["facet", "spec", "resolve"]
    classic_plot_dict = {
        k: v for k, v in classic_plot_dict.items() if k in keys_to_test
    }

    self.assertDictEqual(classic_plot_dict, expected_classic_dict_one_vline)

  def testClassicPlot_one_vline_integer_index(self):
    # Create plot object and use to_dict() to convert the plot components into
    # a more easily queried dict.
    classic_plot_dict = ci.plot(self.ci_data_integer_index).to_dict()

    # Check the important elements: that facets are mapped to the correct
    # variable, the plot layers (lines, bands, etc) have the correct specs, and
    # the scale resolutions are the same.
    keys_to_test = ["facet", "spec", "resolve"]
    classic_plot_dict = {
        k: v for k, v in classic_plot_dict.items() if k in keys_to_test
    }
    self.assertDictEqual(classic_plot_dict,
                         expected_classic_dict_one_vline_integer_index)

  def testClassicPlot_two_vlines(self):
    # Create plot object and use to_dict() to convert the plot components into
    # a more easily queried dict.
    classic_plot_dict = ci.plot(self.ci_data_2).to_dict()

    # Check the important elements: that facets are mapped to the correct
    # variable, the plot layers (lines, bands, etc) have the correct specs, and
    # the scale resolutions are the same.
    keys_to_test = ["facet", "spec", "resolve"]
    classic_plot_dict = {
        k: v for k, v in classic_plot_dict.items() if k in keys_to_test
    }
    self.assertDictEqual(classic_plot_dict, expected_classic_dict_two_vlines)

  def testClassicPlot_four_vlines(self):
    # Create plot object and use to_dict() to convert the plot components into
    # a more easily queried dict.
    classic_plot_dict = ci.plot(self.ci_data_4).to_dict()

    # Check the important elements: that facets are mapped to the correct
    # variable, the plot layers (lines, bands, etc) have the correct specs, and
    # the scale resolutions are the same.
    keys_to_test = ["facet", "spec", "resolve"]
    classic_plot_dict = {
        k: v for k, v in classic_plot_dict.items() if k in keys_to_test
    }
    self.assertDictEqual(classic_plot_dict, expected_classic_dict_four_vlines)

  def testInteractivePlot(self):

    # Create plot object and use to_dict() to convert the plot components into
    # a more easily queried dict.
    interactive_plot_dict = ci.plot(self.ci_data_2, static_plot=False).to_dict()

    # Extract the relevant subdictionaries. Remove the `data` elements since
    # we don't need to check those.
    top_plot_dict = interactive_plot_dict["hconcat"][0]["vconcat"][0]
    bot_plot_dict = interactive_plot_dict["hconcat"][0]["vconcat"][1]
    legend_dict = interactive_plot_dict["hconcat"][1]
    del top_plot_dict["data"]
    del bot_plot_dict["data"]
    del legend_dict["data"]

    # Check that: 1) facets are mapped to the correct variable, 2) plot layers
    # (lines, bands, etc) have the correct specs for both the top panel and the
    # bottom three dynamic plot panels, and 3) the interactive legend is
    # correctly specified.
    self.assertDictEqual(top_plot_dict, expected_top_dict)
    self.assertDictEqual(bot_plot_dict, expected_bot_dict)
    self.assertDictEqual(legend_dict, expected_legend_dict)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019-2023 The TFP CausalImpact Authors
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for causalimpact_lib.py."""

import os

from absl.testing import absltest
from absl.testing import parameterized
import causalimpact as ci
from causalimpact import causalimpact_lib
from causalimpact import data as cid
import numpy as np
import pandas as pd
from statsmodels.tsa.arima_process import ArmaProcess
import tensorflow as tf


CURR_PATH = os.path.dirname(__file__)
TEST_DATA = os.path.join(CURR_PATH, "testdata", "data.csv")


def _create_test_data(treat_amt, treat_index, num_timesteps=100):
  ar = np.r_[1, 0.9]
  ma = np.array([1])
  arma_process = ArmaProcess(ar, ma)
  x = 100 + arma_process.generate_sample(nsample=num_timesteps)
  y = 1.2 * x + np.random.normal(size=(num_timesteps))
  test_data = pd.DataFrame({"y": y, "x": x}, columns=["y", "x"])
  date_index = pd.date_range("2018-01-01", periods=len(y), freq="D")
  test_data.index = date_index
  test_data.loc[test_data.index > date_index[treat_index], "y"] += treat_amt
  return test_data


def _create_impact_data():
  # Make some fake impact data.
  n_time_points = 100
  n_samples = 10
  treat_start_index = 50
  post_period_length = n_time_points - treat_start_index
  treatment_effect = 5.
  cumulative_treatment_effect = treatment_effect * post_period_length
  index = pd.date_range("2018-01-01", periods=n_time_points, freq="D")
  treatment_start = index[treat_start_index]
  pre_period = (index[0], index[treat_start_index - 1])
  post_period = (index[treat_start_index], index[-1])
  colnames = ["sample_" + str(i) for i in range(n_samples)]

  # Make fake data that is just the sequence 0:(n_time_points-1), with an
  # additive treatment effect of 5.
  observed_ts_full = pd.Series(
      np.arange(n_time_points).astype(float), index=index)
  observed_ts_full[treat_start_index:] += treatment_effect

  # The fake "sampled" posterior trajectories are just the sequence
  # 0:(n_time_points-1), i.e. the observed data minus the treatment effect of
  # 5. Note that the posterior trajectories are identical to each other.
  posterior_trajectories = pd.DataFrame(
      pd.concat([pd.Series(np.arange(n_time_points))] * n_samples, axis=1))
  posterior_trajectories.columns = colnames
  posterior_trajectories.index = index

  # The point effect trajectories are just zero in the pre-period and the
  # treatment effect in the post-period.
  one_point_trajectory = pd.Series(np.repeat([0, treatment_effect], 50))
  one_point_trajectory.index = index
  point_trajectories = pd.DataFrame(
      pd.concat([one_point_trajectory] * n_samples, axis=1))
  point_trajectories.columns = colnames

  # Cumulative effect trajectories are the cumulative sum of the expected point
  # trajectories over the post-period (and zero in the pre-period).
  cumulative_trajectories = point_trajectories.cumsum(axis=0)

  # Combine all trajectories.
  trajectory_dict = {
      "predictions": posterior_trajectories,
      "point_effects": point_trajectories,
      "cumulative_effects": cumulative_trajectories
  }

  # Since the fake trajectories are identical, the quantiles, means, and
  # will be equal to a single trajectory. Here we
  # create a dataframe with the summary statistics for the posterior, point, and
  # cumulative trajectories.
  sample_cols = ["sample_1", "sample_2", "sample_3"]
  summary_stats = ["mean", "lower", "upper"]
  posterior_trajectory_summary = posterior_trajectories[sample_cols].copy()
  posterior_trajectory_summary.columns = [
      "posterior_" + stat for stat in summary_stats
  ]
  point_trajectory_summary = point_trajectories[sample_cols].copy()
  point_trajectory_summary.columns = [
      "point_effects_" + stat for stat in summary_stats
  ]
  cumulative_trajectory_summary = cumulative_trajectories[sample_cols].copy()
  cumulative_trajectory_summary.columns = [
      "cumulative_effects_" + stat for stat in summary_stats
  ]

  # Join all trajectory summaries on the time index.
  trajectory_summary = pd.concat([
      observed_ts_full, posterior_trajectory_summary, point_trajectory_summary,
      cumulative_trajectory_summary
  ],
                                 axis=1)
  trajectory_summary.rename(columns={0: "observed"}, inplace=True)

  # For the impact summary table, we need to compute the expected post-period
  # predictions, which is just the sequence treat_start_index:(n_time_points-1),
  # i.e. the observed data minus the treatment effect.
  expected_pred = np.arange(start=treat_start_index, stop=n_time_points)

  # Compute impact summary. Because each "sampled" trajectory for this fake data
  # is identical, the lower/upper values for each quantity are identical to
  # their corresponding point estimates and all standard deviations are zero.
  expected_summary_dict = {
      "actual": {
          "average": observed_ts_full[treat_start_index:].mean(),
          "cumulative": observed_ts_full[treat_start_index:].sum()
      },
      "predicted": {
          "average": expected_pred.mean(),
          "cumulative": expected_pred.sum()
      },
      "predicted_lower": {
          "average": expected_pred.mean(),
          "cumulative": expected_pred.sum()
      },
      "predicted_upper": {
          "average": expected_pred.mean(),
          "cumulative": expected_pred.sum()
      },
      "predicted_sd": {
          "average": 0.,
          "cumulative": 0.
      },
      "abs_effect": {
          "average": treatment_effect,
          "cumulative": cumulative_treatment_effect
      },
      "abs_effect_lower": {
          "average": treatment_effect,
          "cumulative": cumulative_treatment_effect
      },
      "abs_effect_upper": {
          "average": treatment_effect,
          "cumulative": cumulative_treatment_effect
      },
      "abs_effect_sd": {
          "average": 0.,
          "cumulative": 0.
      },
      "rel_effect": {
          "average": treatment_effect / expected_pred.mean(),
          "cumulative": (cumulative_treatment_effect) / expected_pred.sum()
      },
      "rel_effect_lower": {
          "average": treatment_effect / expected_pred.mean(),
          "cumulative": (cumulative_treatment_effect) / expected_pred.sum()
      },
      "rel_effect_upper": {
          "average": treatment_effect / expected_pred.mean(),
          "cumulative": (cumulative_treatment_effect) / expected_pred.sum()
      },
      "rel_effect_sd": {
          "average": 0.,
          "cumulative": 0.
      }
  }
  expected_summary = pd.DataFrame(expected_summary_dict)

  expected_summary["p_value"] = 1. / (n_samples + 1)
  expected_summary["alpha"] = 0.05

  return {
      "treatment_start": treatment_start,
      "pre_period": pre_period,
      "post_period": post_period,
      "observed_ts_full": observed_ts_full,
      "observed_ts_post": observed_ts_full[treat_start_index:],
      "posterior_summary": posterior_trajectory_summary,
      "trajectory_dict": trajectory_dict,
      "trajectory_summary": trajectory_summary,
      "expected_summary": expected_summary
  }


class CausalImpactTest(parameterized.TestCase):

  @classmethod
  def setUpClass(cls):
    super(CausalImpactTest, cls).setUpClass()

    # Read in sample data and set column "t" as the index
    with open(TEST_DATA, "r") as fl:
      df = pd.read_csv(fl)
    df = df.set_index(pd.to_datetime(df["t"]))
    df.drop(columns=["t"], inplace=True)
    # Explicitly set the frequency to match the input data, which is required.
    df.index.freq = "10s"
    df.y[[1, 3, 7]] = np.nan  # Set a few pre-period observations to be missing.
    cls.data = df
    treatment_start_index = 60
    cls.pre_period = (df.index[0], df.index[treatment_start_index - 1])
    cls.post_period = (df.index[treatment_start_index], df.index[-1])
    cls.treatment_start = cls.post_period[0]

  def testShortestPeriodAfterPrePeriod(self):
    ci_analysis = ci.fit_causalimpact(
        data=self.data,
        pre_period=(self.data.index[0], self.data.index[-2]),
        post_period=(self.data.index[-1], self.data.index[-1]),
        inference_options=ci.InferenceOptions(num_results=10),
        seed=(1, 2))
    self.assertIsNotNone(ci_analysis)

  def testUnexpectedKwargsRaisesAnError(self):
    """kwargs are accepted for experimental args, but typos should be caught."""
    with self.assertRaises(TypeError):
      ci.fit_causalimpact(
          some_unknown_arg=3,
          data=self.data,
          pre_period=(self.data.index[0], self.data.index[-2]),
          post_period=(self.data.index[-1], self.data.index[-1]),
          inference_options=ci.InferenceOptions(num_results=10),
          seed=(1, 2))

  @parameterized.named_parameters(
      {
          "testcase_name": "0.01",
          "prior_level_sd": 0.01,
      }, {
          "testcase_name": "0.1",
          "prior_level_sd": 0.1,
      }, {
          "testcase_name": "0.5",
          "prior_level_sd": 0.5,
      })
  def testPriorLevelSdIsUsed(self, prior_level_sd):
    seed = (0, 0)
    treatment_start = 20
    data = self.data
    ci_analysis = ci.fit_causalimpact(
        data=data,
        pre_period=(data.index[0], data.index[treatment_start - 1]),
        post_period=(data.index[treatment_start], data.index[-1]),
        inference_options=ci.InferenceOptions(
            num_results=100, num_warmup_steps=100),
        model_options=ci.ModelOptions(prior_level_sd=prior_level_sd),
        seed=seed)
    # It is a little surprising this passes -- it may be because we are Learning
    # on a small amount of data, but it also may be because the InverseGamma
    # prior is too strong.
    np.testing.assert_allclose(
        np.mean(ci_analysis.posterior_samples.level_scale),
        prior_level_sd,
        atol=0.2 * prior_level_sd)

  def testModelTrainingNoDatetimeIndexSucceeds(self):
    seed = (0, 0)
    data = self.data.copy()
    data.index = np.arange(data.shape[0])
    treatment_start = 20
    ci_analysis = ci.fit_causalimpact(
        data=data,
        pre_period=(data.index[0], data.index[treatment_start - 1]),
        post_period=(data.index[treatment_start], data.index[-1]),
        inference_options=ci.InferenceOptions(num_results=10),
        seed=seed)
    self.assertIsNotNone(ci_analysis)

  def testInterceptIsIncluded(self):
    seed = (0, 0)
    ci_analysis = ci.fit_causalimpact(
        data=self.data,
        pre_period=self.pre_period,
        post_period=self.post_period,
        inference_options=ci.InferenceOptions(num_results=10),
        seed=seed)
    # TEST_DATA.txt has 2 features, plus one intercept
    self.assertEqual(ci_analysis.posterior_samples.weights.shape[-1], 3)

  def testModelTrainingNoCovariates(self):
    ci_data = cid.CausalImpactData(
        self.data["y"],
        pre_intervention_period=self.pre_period,
        post_intervention_period=self.post_period,
        dtype=tf.float64)
    sts_model = causalimpact_lib._build_default_gibbs_model(
        ci_data.normalized_whole_period_features,
        ci_data.pre_intervention_target_ts,
        outcome_sd=tf.constant(1., dtype=tf.float64),
        level_scale=tf.constant(0.01, dtype=tf.float64),
        dtype=tf.float64,
        seasons=[])
    model_params = [p.name for p in sts_model.parameters]
    # TODO(colcarroll,jburnim): Consider how to make these tests more robust,
    # as they will currently break on name changes in the underlying libraries.
    self.assertIn("observation_noise_scale", model_params)
    self.assertTrue(
        any(["LocalLevel/_level_scale" in param for param in model_params]) or
        # Name used by GibbsSampler.
        any(["local_level/_level_scale" in param for param in model_params]))

  def testModelTrainingWithCovariates(self):
    seed = (1, 1)
    ci_data = cid.CausalImpactData(
        self.data,
        pre_intervention_period=self.pre_period,
        post_intervention_period=self.post_period,
        dtype=tf.float32)
    posterior_samples, *_ = causalimpact_lib._train_causalimpact_sts(
        ci_data=ci_data,
        seed=seed,
        num_warmup_steps=100,
        num_results=10,
        prior_level_sd=0.01,
        dtype=tf.float32,
        seasons=[])

    self.assertFalse(np.any(np.isnan(posterior_samples.level)))
    self.assertTrue(np.all(posterior_samples.observation_noise_scale <= 1.2))
    self.assertTrue(np.all(posterior_samples.level_scale <= 1))
    self.assertFalse(np.any(np.isnan(posterior_samples.weights)))

  def testPredictionDims_NoCovars(self):
    num_results = 10
    impact = ci.fit_causalimpact(
        self.data["y"],
        inference_options=ci.InferenceOptions(num_results=num_results),
        pre_period=self.pre_period,
        post_period=self.post_period,
    )
    self.assertIsNotNone(impact.posterior_samples.observation_noise_scale)
    self.assertIsNotNone(impact.posterior_samples.level_scale)
    self.assertIsNotNone(impact.posterior_samples.level)
    # There are no covariates, expect this to not be set.
    self.assertIsNone(impact.posterior_samples.weights)

    # Check that time indices are correct.
    self.assertTrue(impact.series.index.equals(self.data.index))

    # Check that we got the right number of samples.
    self.assertEqual(impact.posterior_samples.observation_noise_scale.shape[0],
                     num_results)

  def testPredictionDims_WithCovars(self):
    num_results = 10
    impact = ci.fit_causalimpact(
        self.data,
        inference_options=ci.InferenceOptions(num_results=num_results),
        pre_period=self.pre_period,
        post_period=self.post_period,
    )

    # Check that time indices are correct.
    self.assertTrue(impact.series.index.equals(self.data.index))

    # Check that we got the right number of samples.
    self.assertEqual(impact.posterior_samples.observation_noise_scale.shape[0],
                     num_results)
    # Confirm that for a low dimensional example, the weights are never 0,
    # since the default prior sets the nonzero probability to 1 for
    # fewer than 3 dimensions (in which case overfitting is not a great fear).
    self.assertEqual((impact.posterior_samples.weights.numpy() == 0).sum(), 0)

  def testComputeImpactTrajectories(self):
    """Test for _compute_impact_trajectories()."""
    # Load test data.
    test_impact_data = _create_impact_data()
    expected_trajectories = test_impact_data["trajectory_dict"]

    # Pull out the posterior trajectories. They are an input to
    # _compute_impact_trajectories() and returned unmodified in its output
    # (just renamed to "predictions").
    posterior_trajectories = expected_trajectories["predictions"]

    trajectory_dict = causalimpact_lib._compute_impact_trajectories(
        posterior_trajectories=posterior_trajectories,
        observed_ts_full=test_impact_data["observed_ts_full"],
        treatment_start=test_impact_data["treatment_start"])

    expected_keys = ["predictions", "point_effects", "cumulative_effects"]
    for k in expected_keys:
      self.assertTrue(trajectory_dict[k].equals(expected_trajectories[k]))

  def testComputeImpactEstimates(self):
    """Test for _compute_impact_estimates()."""
    # Load test data.
    test_impact_data = _create_impact_data()
    expected_trajectories = test_impact_data["trajectory_dict"]
    expected_impact_estimates = test_impact_data["trajectory_summary"]

    # Compute trajectories of point, cumulative, and relative effects using
    # test data.
    trajectory_dict = causalimpact_lib._compute_impact_trajectories(
        posterior_trajectories=expected_trajectories["predictions"],
        observed_ts_full=test_impact_data["observed_ts_full"],
        treatment_start=test_impact_data["treatment_start"])
    impact_estimates = causalimpact_lib._compute_impact_estimates(
        posterior_trajectory_summary=test_impact_data["posterior_summary"],
        trajectory_dict=trajectory_dict,
        observed_ts_full=test_impact_data["observed_ts_full"],
        ci_data=cid.CausalImpactData(
            data=test_impact_data["observed_ts_full"],
            pre_intervention_period=test_impact_data["pre_period"],
            post_intervention_period=test_impact_data["post_period"],
        ),
        quantiles=(0.025, 0.975))

    # Make sure the impact estimates are what we expect. Use column names in
    # expected_impact_estimates to make sure the columns are in the same order
    # between the two dataframes.
    col_names = expected_impact_estimates.columns
    pd.testing.assert_frame_equal(impact_estimates[col_names],
                                  expected_impact_estimates)

  def testComputeImpactSummary(self):
    """Test for _compute_summary()."""

    # Load test data.
    test_impact_data = _create_impact_data()
    expected_trajectories = test_impact_data["trajectory_dict"]
    posterior_trajectory_summary = (test_impact_data["trajectory_summary"])
    expected_summary = test_impact_data["expected_summary"]

    trajectory_dict = causalimpact_lib._compute_impact_trajectories(
        posterior_trajectories=expected_trajectories["predictions"],
        observed_ts_full=test_impact_data["observed_ts_full"],
        treatment_start=test_impact_data["treatment_start"])

    # Calculate impact summary using test data.
    summary = causalimpact_lib._compute_summary(
        posterior_trajectory_summary=posterior_trajectory_summary,
        trajectory_dict=trajectory_dict,
        observed_ts_post=test_impact_data["observed_ts_post"],
        post_period=test_impact_data["post_period"],
        quantiles=(0.025, 0.975),
        alpha=0.05)

    # Make sure the columns and indices are what we expect.
    self.assertSetEqual(set(summary.columns), set(expected_summary.columns))
    self.assertTrue(summary.index.equals(pd.Index(["average", "cumulative"])))

    # Make sure the values are what we expect.
    pd.testing.assert_frame_equal(expected_summary, summary)

  @parameterized.named_parameters(
      {
          "testcase_name": "TupleInts",
          "seed": (13, 37),
      }, {
          "testcase_name": "Int",
          "seed": 14
      })
  def testEvaluate(self, seed):
    """Test for evaluate().

    The workhorse functions of evaluate() are tested in the two previous tests,
    testComputeImpactEstimates() and testComputeImpactSummary(). Here we
    simply test that evaluate() creates the expected attributes (series and
    summary).

    Args:
     seed: Seed to train with.
    """
    treat_index = 50
    treat_amt = 5
    test_data = _create_test_data(treat_amt, treat_index)
    impact = ci.fit_causalimpact(
        test_data.copy(),
        pre_period=(test_data.index[0], test_data.index[treat_index - 1]),
        post_period=(test_data.index[treat_index], test_data.index[-1]),
        inference_options=ci.InferenceOptions(num_results=10),
        seed=seed)
    self.assertIsInstance(impact.series, pd.DataFrame)
    self.assertIsInstance(impact.summary, pd.DataFrame)

    # Check that `evaluate` is deterministic (when using a stateless seed).
    new_impact = ci.fit_causalimpact(
        test_data,
        pre_period=(test_data.index[0], test_data.index[treat_index - 1]),
        post_period=(test_data.index[treat_index], test_data.index[-1]),
        inference_options=ci.InferenceOptions(num_results=10),
        seed=seed)

    pd.testing.assert_frame_equal(impact.series, new_impact.series)
    pd.testing.assert_frame_equal(impact.summary, new_impact.summary)

  @parameterized.named_parameters(
      {
          "testcase_name": "NoTimeAfterPostPeriod",
          "time_after_post_period": False,
      }, {
          "testcase_name": "TimeAfterPostPeriod",
          "time_after_post_period": True
      })
  def testSummary(self, time_after_post_period):
    """Verifies that time after the post period does not change the summary."""
    treat_index = 50
    treat_amt = 5
    timesteps_before_post_period_ends = 100
    # In the extra time case, just add an extra set of points but do NOT
    # move the post-period.
    num_timesteps = timesteps_before_post_period_ends + (
        50 if time_after_post_period else 0)
    test_data = _create_test_data(
        treat_amt, treat_index, num_timesteps=num_timesteps)
    pre_period = (test_data.index[0], test_data.index[treat_index - 1])
    post_period = (test_data.index[treat_index],
                   test_data.index[timesteps_before_post_period_ends - 1])
    impact = ci.fit_causalimpact(
        test_data.copy(),
        pre_period=pre_period,
        post_period=post_period,
        inference_options=ci.InferenceOptions(num_results=10),
        seed=0)
    # This confirms that the total cumulative effect is approximately correct.
    # If we did not respect the end of post periods, it would be double.
    np.testing.assert_allclose(
        impact.summary.loc["cumulative", "abs_effect"], 250, rtol=0.2)

  def testNonAlignedStartTime(self):
    """Verifies that pre/post periods do not need to be aligned with the data."""
    n_time_steps = 100
    treat_start = 50
    true_effect = 5.
    y = np.random.normal(size=n_time_steps, scale=0.0001)
    y[treat_start:] += true_effect
    date_index = pd.date_range("2018-01-07", periods=n_time_steps, freq="W")
    test_data = pd.DataFrame({"y": y}, index=date_index)

    analysis = ci.fit_causalimpact(
        test_data,
        # Choose dates not aligned to 2018-01-07 on a weekly cadence.
        pre_period=("2018-01-10", "2018-01-30"),
        post_period=("2018-02-02", "2018-02-23"),
        inference_options=ci.InferenceOptions(num_results=10))
    # Verify that we have no information set before post period starts,
    # rounding up.
    self.assertEqual(
        # TODO(colcarroll,jburnim): This should be NaN, not 0.
        0,
        analysis.series.loc[pd.to_datetime("2018-01-28"),
                            "cumulative_effects_mean"])
    self.assertNotEqual(
        0, analysis.series.loc[pd.to_datetime("2018-02-04"),
                               "cumulative_effects_mean"])

  def testGapBetweenPreAndPostPeriod(self):
    """Verifies that pre/post periods do not need to be aligned with the data."""
    n_time_steps = 20
    treat_start = 50
    true_effect = 5.
    y = np.random.normal(size=n_time_steps, scale=0.0001)
    y[treat_start:] += true_effect
    date_index = pd.date_range("2022-01-07", periods=n_time_steps, freq="D")
    data_period = (date_index[0], date_index[-1])
    test_data = pd.DataFrame({"y": y}, index=date_index)

    before_pre_period = (data_period[0], "2022-01-08")
    pre_period = ("2022-01-09", "2022-01-12")
    inbetween_period = ("2022-01-13", "2022-01-14")
    # Post period finishes before all timesteps provided, to test that
    # behaviour.
    post_period = ("2022-01-15", "2022-01-20")
    after_post_period = ("2022-01-21", data_period[1])

    analysis = ci.fit_causalimpact(
        test_data,
        pre_period=pre_period,
        post_period=post_period,
        inference_options=ci.InferenceOptions(num_results=10))
    series = analysis.series
    # Drop the non-number values for easier testing.
    series.drop([
        "pre_period_start", "pre_period_end", "post_period_start",
        "post_period_end"
    ],
                inplace=True,
                axis="columns")
    with self.subTest("IndexMathes"):
      # Expect data for all points, even outside the pre- and post-period range.
      pd.testing.assert_index_equal(series.index, date_index)

    with self.subTest("EntirePeriod"):
      # Verify observed is returned, even outside of the pre- and post-period
      # range.
      self.assertTrue(series.loc[data_period[0]:data_period[1],
                                 ["observed"]].notna().all(axis=None))

    with self.subTest("BeforePreperiod"):
      # Only observed should be set before pre-period.
      self.assertTrue(
          series.loc[before_pre_period[0]:before_pre_period[1],
                     series.columns.difference(["observed"])].isna().all(
                         axis=None))

    with self.subTest("Pre-period"):
      # Verify posterior and point effects have values.
      self.assertTrue(series.loc[pre_period[0]:pre_period[1], [
          "posterior_mean", "posterior_lower", "posterior_upper",
          "point_effects_mean", "point_effects_lower", "point_effects_upper"
      ]].notna().all(axis=None))
      # Verify cumulative values are zero.
      self.assertTrue(series.loc[pre_period[0]:pre_period[1], [
          "cumulative_effects_mean", "cumulative_effects_lower",
          "cumulative_effects_upper"
      ]].eq(0).all(axis=None))

    with self.subTest("Inbetween-period"):
      # Verify posterior values are returned.
      self.assertTrue(series.loc[
          inbetween_period[0]:inbetween_period[1],
          ["posterior_mean", "posterior_lower", "posterior_upper"]].notna().all(
              axis=None))
      # Verify no point and cumulative values.
      self.assertTrue(series.loc[inbetween_period[0]:inbetween_period[1], [
          "point_effects_mean", "point_effects_lower", "point_effects_upper",
          "cumulative_effects_mean", "cumulative_effects_lower",
          "cumulative_effects_upper"
      ]].isna().all(axis=None))

    with self.subTest("Post-period"):
      # Verify that all values are returned.
      self.assertTrue(
          series.loc[post_period[0]:post_period[1]].notna().all(axis=None))

    with self.subTest("AfterPostperiod"):
      # Only the observed and posterior should be set after post-period.
      expected_columns = [
          "observed", "posterior_mean", "posterior_lower", "posterior_upper"
      ]
      self.assertTrue(
          series.loc[after_post_period[0]:after_post_period[1],
                     series.columns.difference(expected_columns)].isna().all(
                         axis=None))
      self.assertTrue(series.loc[after_post_period[0]:after_post_period[1],
                                 expected_columns].notna().all(axis=None))

  @parameterized.named_parameters(
      {
          "testcase_name": "float32",
          "dtype": tf.float32
      }, {
          "testcase_name": "float64",
          "dtype": tf.float64
      })
  def testNumericImpactValues(self,
                              dtype=tf.float32):
    """Test for numeric values of evaluate().

    This test uses simulated data with a large treatment effect and a large
    number of MCMC draws to check that the numeric values of the estimated
    effects and their uncertainty intervals are close to what we'd expect.

    Args:
      dtype: The dtype to use throughout computation.
    """
    n_time_steps = 100
    treat_start = 50
    true_effect = 5.
    y = np.random.normal(size=n_time_steps, scale=0.0001)
    y[treat_start:] += true_effect
    # TODO(colcarroll): Add a version of this test that does not standardize
    # data and by adding (say) 100 to `y`, then asserting that
    # `mod.series[0, posterior_lower] < y[0]`, and that
    # y[0] < mod.series[0, posterior_upper]`
    date_index = pd.date_range("2018-01-01", periods=n_time_steps, freq="D")
    test_data = pd.DataFrame({"y": y}, index=date_index)
    impact = ci.fit_causalimpact(
        test_data,
        pre_period=(test_data.index[0], test_data.index[treat_start - 1]),
        post_period=(test_data.index[treat_start], test_data.index[-1]),
        inference_options=ci.InferenceOptions(num_results=1000),
        data_options=ci.DataOptions(dtype=dtype))
    summary = impact.summary
    # The true absolute effects, as average and sum over post period.
    true_abs_effects = (true_effect, true_effect * (n_time_steps - treat_start))
    np.testing.assert_allclose(
        summary["abs_effect"], true_abs_effects, rtol=0.001, atol=0.001)

    # See that interval widths (relative to absolute effects) are about 0.
    relative_interval_widths = (
        (summary["abs_effect_upper"] - summary["abs_effect_lower"]) /
        summary["abs_effect"])
    self.assertLessEqual(relative_interval_widths["average"], 0.01)
    self.assertLessEqual(relative_interval_widths["cumulative"], 0.01)

  def testNumericImpactValuesWithSeasonality(self):
    """Verifies that seasonality assists with inference.

    This is not intended to do a detailed analysis of the inference, just that
    there is difference when seasonal effects are modeled when the data
    does have a true underlying effect.
    """
    dtype = tf.float32
    n_time_steps = 300
    treat_start = 290
    true_effect = 2.5
    every_five_effect = [
        [8., 8., 4., 3., -4.][x % 5] for x in range(n_time_steps)
    ]
    every_seven_effect = [
        10 * [1., 4., 5., 2., -1., -2., -3.][x % 7] for x in range(n_time_steps)
    ]
    every_eight_effect = [
        [1., 1., 3., 3., 4.5, 2.0, -7., 0.][x % 8] for x in range(n_time_steps)
    ]
    y = (
        np.random.normal(size=n_time_steps, scale=0.4) + every_seven_effect +
        every_five_effect + every_eight_effect)

    y[treat_start:] += true_effect
    date_index = pd.date_range("2018-01-01", periods=n_time_steps, freq="D")
    test_data = pd.DataFrame({"y": y}, index=date_index)
    impact_without_season = ci.fit_causalimpact(
        test_data,
        pre_period=(test_data.index[0], test_data.index[treat_start - 1]),
        post_period=(test_data.index[treat_start], test_data.index[-1]),
        inference_options=ci.InferenceOptions(num_results=1000),
        data_options=ci.DataOptions(dtype=dtype))

    impact_with_season = ci.fit_causalimpact(
        test_data,
        pre_period=(test_data.index[0], test_data.index[treat_start - 1]),
        post_period=(test_data.index[treat_start], test_data.index[-1]),
        inference_options=ci.InferenceOptions(num_results=1000),
        data_options=ci.DataOptions(dtype=dtype),
        model_options=ci.ModelOptions(seasons=[
            ci.Seasons(
                num_seasons=4,
                # Ensure we can handle num_steps_per_season
                # being a tuple.
                num_steps_per_season=(2, 1, 1, 1)),
            ci.Seasons(num_seasons=7),
            ci.Seasons(
                num_seasons=6,
                # Ensure we can handle num_steps_per_season
                # being a nested tuple.
                num_steps_per_season=((2, 2, 1, 1, 1, 1), (2, 2, 1, 1, 1, 1)))
        ]))

    # Estimates of the absolute effect will be very wide when seasonality
    # is not modeled, since the variance is unexplained. When the seasonality
    # is modeled, the estimates will be much tighter.
    self.assertAlmostEqual(
        9.5,
        impact_without_season.summary["abs_effect_sd"]["average"],
        delta=1.)
    self.assertAlmostEqual(
        0.5, impact_with_season.summary["abs_effect_sd"]["average"], delta=0.1)

    self.assertSequenceEqual(
        [1000, 300, 0],
        impact_without_season.posterior_samples.seasonal_levels.shape)
    self.assertSequenceEqual(
        [1000, 300, 3],
        impact_with_season.posterior_samples.seasonal_levels.shape)


class _CausalImpactBaseTest(absltest.TestCase):

  def check_all_functions(self, impact):
    self.assertIn("Posterior Inference", ci.summary(impact))
    self.assertIn("Posterior Inference",
                  ci.summary(impact, output_format="summary"))
    self.assertIn("Analysis report", ci.summary(impact, output_format="report"))
    self.assertIn("Posterior Inference", ci.summary(impact, "summary"))
    self.assertIn("Analysis report", ci.summary(impact, "report"))
    with self.assertRaises(ValueError):
      ci.summary(impact, output_format="foo")

    self.assertIsNotNone(ci.plot(impact))


class TestDataFormats(_CausalImpactBaseTest):

  def test_missing_input(self):
    with self.assertRaises(TypeError):
      ci.fit_causalimpact()

  def test_healthy_input(self):
    data = pd.DataFrame({
        "y": np.random.randn(200),
        "x1": np.random.randn(200),
        "x2": np.random.randn(200)
    })
    impact = ci.fit_causalimpact(
        data,
        pre_period=(0, 100),
        post_period=(101, 199),
        inference_options=ci.InferenceOptions(num_results=10))
    self.assertEqual(data.shape[0], impact.series.shape[0])
    self.check_all_functions(impact)


class TestPreAndPostPeriod(_CausalImpactBaseTest):

  def test_missing_pre_period_input(self):
    data = pd.DataFrame({
        "y": np.random.randn(200),
        "x1": np.random.randn(200),
        "x2": np.random.randn(200)
    })
    data.y.iloc[2:5] = np.nan
    impact = ci.fit_causalimpact(
        data,
        pre_period=(0, 100),
        post_period=(101, 199),
        inference_options=ci.InferenceOptions(num_results=10))
    self.assertEqual(data.shape[0], impact.series.shape[0])
    self.check_all_functions(impact)

    # Verify that the indices that have NaN outputs have NaN for everything
    # other than point predictions.
    self.assertTrue(
        np.isnan(impact.series.iloc[2:5][impact.series.columns.difference([
            "observed",
            "posterior_mean",
            "posterior_lower",
            "posterior_upper",
            "time",
            "pre_period_start",
            "pre_period_end",
            "post_period_start",
            "post_period_end",
        ])]).all(
            # Reduce across both axes.
            axis=None))

if __name__ == "__main__":
  absltest.main()



# Copyright 2019-2023 The TFP CausalImpact Authors
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFP-based CausalImpact implementation via fit_causalimpact."""

import dataclasses
import logging
import math
from typing import Dict, List, Optional, Tuple, Union

from causalimpact import posterior_processing
import causalimpact.data as cid
from causalimpact.indices import InputDateType
from causalimpact.indices import OutputDateType
from causalimpact.indices import OutputPeriodType

import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability.python.experimental.distributions import MultivariateNormalPrecisionFactorLinearOperator
from tensorflow_probability.python.experimental.sts_gibbs import gibbs_sampler
from tensorflow_probability.python.internal import prefer_static as ps

tfb = tfp.bijectors
tfd = tfp.distributions

TensorLike = tf.types.experimental.TensorLike
_SeedType = Union[int, Tuple[int, int], TensorLike]


@dataclasses.dataclass
class CausalImpactPosteriorSamples:
    """Results for samples of the latents of CausalImpact."""
    # Shape is [num_samples].
    observation_noise_scale: tf.Tensor
    # Shape is [num_samples].
    level_scale: Optional[tf.Tensor]
    # Shape is [num_samples].
    level: Optional[tf.Tensor]
    # Shape is [num_samples, num_covariates + 1 (for intercept)].
    weights: Optional[tf.Tensor]
    # Shape is [num_samples, num_seasonal_effects]
    seasonal_drift_scales: Optional[tf.Tensor]
    # Shape is [num_samples, num_seasonal_effects]
    seasonal_levels: Optional[tf.Tensor]


@dataclasses.dataclass
class CausalImpactAnalysis:
    """Results of a CausalImpact analysis.

    Holds both the dataframe outputs of `fit_causalimpact`, as well as the full
    posterior of draws from the structural time series model that was fit. See
    `build_model_for_gibbs_fitting` in the module
    `tensorflow_probability.experimental.sts_gibbs.gibbs_sampler` for details of
    the model parameters.

    Attributes:
      series: DataFrame describing the posterior of the causal impact of the
        intervention at each step. This contains an index with the same
        range as the input data (i.e. not just constrained to
        [pre-period start, post-period end]) with columns:

          observed: The observed value.
          posterior_mean: The mean of the predicted value.
          posterior_lower: The lower bound of the (equal-tailed) credible interval
            for the counterfactual forecast of the observed value.
          posterior_upper: The upper bound of the (equal-tailed) credible interval
            for the counterfactual forecast of the observed value.
          point_effects_mean: The posterior mean of the difference between the
            observed and predicted value.
          point_effects_lower: The lower bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval for the difference between the
            observed and the predicted value.
          point_effects_upper: The upper bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval for the difference between the
            observed and the predicted value.
          cumulative_effects_mean: Starting from the beginning of the post-period,
            the posterior mean of the cumulative difference between the
            observed and predicted value.
          cumulative_effects_lower: The lower bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval for the cumulative difference between
            the observed and the predicted value.
          cumulative_effects_upper: The upper bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval for the cumulative difference between
            the observed and the predicted value.

        Some values may be NaN.
      summary: DataFrame that is a compact representation of `series` that
        summarizes the impact in the post-period. The two rows - labelled
        'average' and 'cumulative' - are computed from the posterior predictive
        samples, and for the first nine columns, the `cumulative` will be a
        constant multiple of the `average`. The columns are:

          actual: The observed value over the post-period.
          predicted: The predicted value over the post-period.
          predicted_lower: Lower bound for the `(1-alpha)`-percentile
            credible interval of the predicted value.
          predicted_upper: Upper bound for the `(1-alpha)`-percentile
            (equal-tailed) credible interval of the predicted value.
          predicted_sd: Standard deviation of the predicted value.
          abs_effect: The posterior mean of the difference between the
            observed and predicted value.
          abs_effect_lower: The lower bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval for the difference between the
            observed and the predicted value.
          abs_effect_upper: The upper bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval for the difference between the
            observed and the predicted value.
          abs_effect_sd: Standard deviation of the difference between the
            aggregation of the actual value and the predicted value.
          rel_effect: Posterior mean of the relative effect
            (observed / predicted - 1). This is the same for both average and
            cumulative.
          rel_effect_lower: The lower bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval of the relative effect.
          rel_effect_upper: The upper bound of the `(1-alpha)`-percentile
            (equal-tailed) credible interval of the relative effect.
          rel_effect_sd: The standard deviation of the relative effect.
          p_value: One-sided p-value that there was a cumulative effect.

      posterior_samples: NamedTuple of Tensorlikes containing samples of latent
        variables. Useful for understanding what the model learned without even
        considering predictions.
    """
    series: pd.DataFrame
    summary: pd.DataFrame
    # Samples of latent variables of the model. Note that in CausalImpact.R,
    # impact$posterior_samples refers to predictive posterior samples, instead
    # of samples of the latents.
    posterior_samples: CausalImpactPosteriorSamples


@dataclasses.dataclass
class DataOptions:
    """Options for working with the input data.

    Attributes:
      outcome_column: String giving the name of the outcome column in `data`. If
        not specified, the first column in `data` is used.
      standardize_data: If covariates and output should be standardized.
      dtype: The dtype to use throughout computation.
    """
    outcome_column: Optional[str] = None
    standardize_data: bool = True
    dtype: tf.dtypes.DType = tf.float32


@dataclasses.dataclass(frozen=True)
class Seasons:
    """Options for modeling a single seasonal effect.

    For instance, modeling day of the week.

    Attributes:
      num_seasons: Number of seasons.
      num_steps_per_season: Number of steps in each season. This may be either an
        int, in which case all seasons have the same length, or a tuple of ints of
        shape `[num_seasons]`, in which seasons have different length, but remain
        constant around different cycles, or a tuple of tuples of ints of shape
        `[num_cycles, num_seasons]`, in which num_steps_per_season for each season
        also varies in different cycle (e.g., a 4 years cycle with leap day).
        Using [num_cycles] allows having the timeseries start in the middle of a
        season.
    """
    num_seasons: int
    num_steps_per_season: Union[int, Tuple[int], Tuple[Tuple[int]]] = 1


@dataclasses.dataclass
class ModelOptions:
    """Options used for the model that is being fit.

    Attributes:
      prior_level_sd: float: Prior standard deviation of the Gaussian random walk
        of the local level model. Expressed in terms of data standard deviations.
        Defaults to 0.01, a typical choice for well-behaved and stable datasets
        with low residual volatility. When in doubt, a safer option is to use 0.1,
        as validated on synthetic data, although this may sometimes give rise to
        unrealistically wide prediction intervals.
      seasons: This supports a list of Seasons, for modeling
        multiple seasons. For instance, for hourly data, there could be
        both an hour-of-the-day (num_seasons=24, num_steps_per_season=1), and
        day-of-the-week effect (num_seasons=7, num_steps_per_season=24) by passing
        in two SeasonalOptions. This is different from a single SeasonalOption
        that would model hour-of-the-day-of-the-week effect (num_seasons=7*24,
        num_steps_per_season=1).
    """
    prior_level_sd: float = 0.01
    seasons: List[Seasons] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class InferenceOptions:
    """Options for how to do inference for the model.

    Attributes:
      num_results: integer number of posterior samples to draw.
      num_warmup_steps: integer number of steps to take before starting to collect
        results. If not set, then 1/9th of num_results will be used.
    """
    num_results: int = 900
    num_warmup_steps: Optional[int] = None

    def __post_init__(self):
        if self.num_warmup_steps is None:
            self.num_warmup_steps = math.ceil(self.num_results / 9)


def fit_causalimpact(data: pd.DataFrame,
                     pre_period: Tuple[InputDateType, InputDateType],
                     post_period: Tuple[InputDateType, InputDateType],
                     alpha: float = 0.05,
                     seed: Optional[_SeedType] = None,
                     data_options: Optional[DataOptions] = None,
                     model_options: Optional[ModelOptions] = None,
                     inference_options: Optional[InferenceOptions] = None,
                     **kwargs) -> CausalImpactAnalysis:
    """Fit a CausalImpact model on provided data.

    Args:
      data: Pandas `DataFrame` containing an outcome time series and optional
        features time series. If a datetime index is used, it is assumed to be of
        a fixed period. Similarly, an integer index assumes all points are
        adjacent (e.g. an index of [2,3,4,6] will assume [4,6] are adjacent for
        training and prediction.
      pre_period: Tuple of datetimes, indices or strings (convertible via
        pd.to_datetime) giving pre-period start/end.
      post_period: Tuple of datetimes, indices or strings (convertible via
        pd.to_datetime) giving post-period start/end.
      alpha: float, used to get confidence level for predictions.
      seed: PRNG seed; see `tensorflow_probability.random.sanitize_seed` for
        details.
      data_options: See DataOptions.
      model_options: See ModelOptions.
      inference_options: See InferenceOptions.
      **kwargs: Only for experimental options. Unsupported keys will raise an
        exception.

    Returns:
      A CausalImpactAnalysis instance summarizing the effect of the intervention.
    """
    # Suppress verbose TensorFlow WARNING and INFO messages, which are not useful
    # for TFP CausalImpact users.
    tf_log_level = tf.get_logger().level
    tf.get_logger().setLevel(logging.ERROR)
    try:
        data_options = data_options if data_options is not None else DataOptions()
        model_options = (model_options if model_options is not None
                         else ModelOptions())
        inference_options = (inference_options if inference_options is not None
                             else InferenceOptions())

        # WARNING: These are implementation details, and have no guarantees of
        # continuing to work or be respected.
        experimental_model = kwargs.pop("experimental_model", None)
        experimental_tf_function_cache_key_addition = kwargs.pop(
            "experimental_tf_function_cache_key_addition", 0)
        if kwargs:
            raise TypeError(f"Received unknown {kwargs=}")

        ci_data = cid.CausalImpactData(
            data=data,
            pre_intervention_period=pre_period,
            post_intervention_period=post_period,
            target_col_name=data_options.outcome_column,
            standardize_data=data_options.standardize_data,
            dtype=data_options.dtype)
        posterior_samples, posterior_means, posterior_trajectories = _train_causalimpact_sts(
            ci_data=ci_data,
            prior_level_sd=model_options.prior_level_sd,
            seed=seed,
            num_results=inference_options.num_results,
            num_warmup_steps=inference_options.num_warmup_steps,
            model=experimental_model,
            dtype=data_options.dtype,
            seasons=model_options.seasons,
            experimental_tf_function_cache_key_addition=experimental_tf_function_cache_key_addition
        )
        series, summary = _compute_impact(
            posterior_means=posterior_means,
            posterior_trajectories=posterior_trajectories,
            ci_data=ci_data,
            alpha=alpha)

        if posterior_samples.seasonal_levels.shape[-1] > 0:
            # If we have k seasonal effects, with num_seasons[0],...,num_seasons[k-1]
            # distinct seasons respectively, then each seasonal effect's latent state
            # has dimension num_seasons[0]-1, ..., num_seasons[k-1]-1 and the shape
            # of posterior_samples.seasonal_levels is
            #   batch_shape [timeseries_length, total_seasonal_latent_dim]
            # where
            #   total_seasonal_latent_dim = num_seasons[0]-1, ..., num_seasons[k-1]-1.
            #
            # And at each timestep, each seasonal effect's contribution to the
            # observed value is the 0-th element of its latent state.  Here we extract
            # these 0-th elements in order to return  the contribution of each
            # seasonal effect at each time step.
            seasonal_levels = []
            index = 0
            for season in model_options.seasons:
                seasonal_levels.append(posterior_samples.seasonal_levels[..., index])
                index += season.num_seasons - 1
            seasonal_levels = tf.stack(seasonal_levels, axis=-1)
        else:
            # If there are no seasonal effects, then we can just use
            # posterior_samples.seasonal_levels, which has the correct shape:
            #     batch_shape + [num_samples, timeseries_length, 0]
            seasonal_levels = posterior_samples.seasonal_levels

        # Translate posterior samples to a CausalImpact-specific object,
        # rather than exposing GibbsSamplerState.
        ci_posterior_samples = CausalImpactPosteriorSamples(
            observation_noise_scale=posterior_samples.observation_noise_scale,
            level_scale=posterior_samples.level_scale,
            level=posterior_samples.level,
            weights=(posterior_samples.weights
                     if posterior_samples.weights.shape[1] > 0 else None),
            seasonal_drift_scales=(
                posterior_samples.seasonal_drift_scales
                if posterior_samples.seasonal_drift_scales.shape[-1] > 0 else None),
            seasonal_levels=seasonal_levels
        )
        return CausalImpactAnalysis(series, summary, ci_posterior_samples)
    finally:
        tf.get_logger().setLevel(tf_log_level)


# Always use graph mode: eager mode is very, very slow. The non-compiled version
# is used with the dynamic Cholesky decompositions, which use a dynamic shape,
# and frustrates the compilation.
@tf.function(autograph=False, jit_compile=False)
def _run_gibbs_sampler(
        sts_model: Optional[tfp.sts.StructuralTimeSeries], outcome_ts: TensorLike,
        outcome_sd: TensorLike, design_matrix: Optional[TensorLike],
        num_results: int, num_warmup_steps: int,
        observation_noise_scale: TensorLike, level_scale: TensorLike,
        seasonal_drift_scales: TensorLike, weights: TensorLike, level: TensorLike,
        slope: TensorLike, seed: TensorLike, dtype, seasons: List[Seasons],
        experimental_tf_function_cache_key_addition: int):  # pylint: disable=unused-argument
    """Fits parameters of the model using Gibbs sampling."""
    if not sts_model:
        sts_model = _build_default_gibbs_model(
            design_matrix=design_matrix,
            outcome_ts=outcome_ts,
            level_scale=level_scale,
            outcome_sd=outcome_sd,
            dtype=dtype,
            seasons=seasons)

    sample_seed, forecast_seed = tfp.random.split_seed(seed)
    posterior_samples = gibbs_sampler.fit_with_gibbs_sampling(
        sts_model,
        outcome_ts,
        num_results=num_results,
        num_warmup_steps=num_warmup_steps,
        initial_state=gibbs_sampler.GibbsSamplerState(
            observation_noise_scale=observation_noise_scale,
            level_scale=level_scale,
            # Model has no slope component.
            slope_scale=tf.zeros([], dtype=dtype),
            weights=weights,
            level=level,
            slope=slope,
            seed=None,
            seasonal_drift_scales=seasonal_drift_scales,
            seasonal_levels=tf.zeros(
                shape=gibbs_sampler.get_seasonal_latents_shape(
                    outcome_ts.time_series, sts_model),
                dtype=dtype)),
        # TODO(colcarroll,jburnim): Move to commented-on module constant.
        default_pseudo_observations=tf.ones([], dtype=dtype) * 0.01,
        seed=sample_seed,
        experimental_use_dynamic_cholesky=True,
        experimental_use_weight_adjustment=True)

    posterior_means, posterior_trajectories = (
        _get_posterior_means_and_trajectories(
            sts_model=sts_model,
            posterior_samples=posterior_samples,
            seed=forecast_seed))
    return posterior_samples, posterior_means, posterior_trajectories


def _build_default_gibbs_model(
        design_matrix: Optional[tf.Tensor],
        outcome_ts: tfp.sts.MaskedTimeSeries,
        level_scale: tf.Tensor,
        outcome_sd: tf.Tensor,
        dtype,
        seasons: List[Seasons],
):
    """A method to build the default STS model.

    The default model is a local level model with observation noise, plus optional
    sparse regression using observed covariates, if given.

    Args:
      design_matrix: Optional Tensor of [timesteps, features] with the covariates.
      outcome_ts: An instance of tfp.sts.MaskedTimeSeries.
      level_scale: tf.Tensor - Initial scale for the local level.
      outcome_sd: Standard deviation of non-nan values in the `observed_ts`. Used
        for scaling parameter defaults.
      dtype: Desired dtype for Gibbs model.
      seasons: An interable of seasonal options for seasonal components to add
        to the model.

    Returns:
      A tfp.sts.StructuralTimeSeries instance.
    """
    local_level_prior_sample_size = tf.constant(32., dtype=dtype)

    level_concentration = tf.cast(local_level_prior_sample_size / 2., dtype=dtype)
    level_variance_prior_scale = level_scale * level_scale * (
            local_level_prior_sample_size / 2.)

    level_variance_prior = tfd.InverseGamma(
        concentration=level_concentration, scale=level_variance_prior_scale)
    level_variance_prior.upper_bound = outcome_sd

    if design_matrix is not None:
        observation_noise_variance_prior = tfd.InverseGamma(
            concentration=tf.constant(25., dtype=dtype),
            scale=tf.math.square(outcome_sd) * tf.constant(5., dtype=dtype))
    else:
        observation_noise_variance_prior = tfd.InverseGamma(
            concentration=tf.constant(0.005, dtype=dtype),
            scale=tf.math.square(outcome_sd) * tf.constant(0.005, dtype=dtype))
    observation_noise_variance_prior.upper_bound = outcome_sd * tf.constant(
        1.2, dtype=dtype)

    if design_matrix is not None:
        design_shape = ps.shape(design_matrix)
        num_outputs = design_shape[-2]
        num_dimensions = design_shape[-1]
        sparse_weights_nonzero_prob = tf.minimum(
            tf.constant(1., dtype=dtype), 3. / num_dimensions)
        x_transpose_x = tf.matmul(design_matrix, design_matrix, transpose_a=True)
        weights_prior_precision = 0.01 * tf.linalg.set_diag(
            0.5 * x_transpose_x, tf.linalg.diag_part(x_transpose_x)) / num_outputs
        # TODO(colcarroll): Remove this cholesky - it is used to instantiate the
        # MVNPFLO below, but later code only uses the precision.
        precision_factor = tf.linalg.cholesky(weights_prior_precision)

        # Note that this prior uses the entire design matrix -- not just the
        # pre-period -- which "cheats" by using future data.
        weights_prior = MultivariateNormalPrecisionFactorLinearOperator(
            precision_factor=tf.linalg.LinearOperatorFullMatrix(precision_factor),
            precision=tf.linalg.LinearOperatorFullMatrix(weights_prior_precision))
    else:
        sparse_weights_nonzero_prob = None
        weights_prior = None

    initial_level_prior = tfd.Normal(
        loc=tf.cast(outcome_ts.time_series[..., 0], dtype=dtype),
        scale=outcome_sd)

    seasonal_components = []
    seasonal_variance_prior = tfd.InverseGamma(
        concentration=0.005, scale=5e-7 * tf.square(outcome_sd))
    seasonal_variance_prior.upper_bound = outcome_sd
    for seasonal_options in seasons:
        seasonal_components.append(
            tfp.sts.Seasonal(
                num_seasons=seasonal_options.num_seasons,
                num_steps_per_season=np.array(
                    seasonal_options.num_steps_per_season),
                allow_drift=True,
                constrain_mean_effect_to_zero=True,
                # TODO(colcarroll,jburnim): If there are multiple seasonal effects,
                # should the prior for the draft scale or initial effect be
                # scaled down?
                drift_scale_prior=tfd.TransformedDistribution(
                    bijector=tfb.Invert(tfb.Square()),
                    distribution=seasonal_variance_prior),
                initial_effect_prior=tfd.Normal(loc=0., scale=outcome_sd)))

    return gibbs_sampler.build_model_for_gibbs_fitting(
        outcome_ts,
        design_matrix=design_matrix,
        weights_prior=weights_prior,
        level_variance_prior=level_variance_prior,
        slope_variance_prior=None,
        observation_noise_variance_prior=observation_noise_variance_prior,
        initial_level_prior=initial_level_prior,
        sparse_weights_nonzero_prob=sparse_weights_nonzero_prob,
        seasonal_components=seasonal_components)


def _train_causalimpact_sts(
        *,
        ci_data: cid.CausalImpactData,
        prior_level_sd,
        seed: _SeedType,
        num_results: int,
        num_warmup_steps: int,
        model: Optional[tfp.sts.StructuralTimeSeries] = None,
        dtype,
        seasons: List[Seasons],
        experimental_tf_function_cache_key_addition: int = 0,
) -> Tuple[gibbs_sampler.GibbsSamplerState, TensorLike, TensorLike]:
    """Structural Time Series (STS) methods for CausalImpact.

    Fits a Tensorflow Probability (TFP) structural time series model.

    Args:
      ci_data: CausalImpact data, used to fit a model.
      prior_level_sd: float: Prior standard deviation of the Gaussian random walk
        of the local level model. Expressed in terms of data standard deviations.
      seed: PRNG seed; see `tensorflow_probability.random.sanitize_seed` for
        details.
      num_results: See `fit_causalimpact`.
      num_warmup_steps: See `fit_causalimpact`.
      model: Optional custom model - `tfp.sts.StructuralTimeSeries`.
      dtype: The dtype to use throughout computation.
      seasons: See `fit_causalimpact`.
      experimental_tf_function_cache_key_addition: See `fit_causalimpact`.

    Returns:
      The model used, posterior samples, posterior means, posterior trajectories
    """
    if isinstance(seed, int):
        # Integers are stateful seeds, thus calling twice with the same seed will
        # return different results. A tuple version is stateless once passed in
        # sanitize_seed.
        seed = (0, seed)
    # While compiled samplers can take non-sanitized seeds, this
    # means our support for non-Tensor seeds will result in cache misses
    # when just the seed changes. Thus sanitize before tracing/compiling.
    seed = tfp.random.sanitize_seed(seed)

    design_matrix = None if ci_data.normalized_whole_period_features is None else tf.convert_to_tensor(
        ci_data.normalized_whole_period_features.values, dtype=dtype)

    # To combine posterior sampling with predictions, instead of just using
    # the pre-period, also use the post-period, but with all values being NaN.
    normalized_after_pre_intervention_period_length = ci_data.normalized_after_pre_intervention_data.shape[0]

    extended_target_ts = tfp.sts.MaskedTimeSeries(
        time_series=tf.concat([
            ci_data.pre_intervention_target_ts.time_series,
            tf.fill(normalized_after_pre_intervention_period_length, tf.constant(
                float("nan"), dtype=dtype))
        ],
            axis=0),
        is_missing=tf.concat([
            ci_data.pre_intervention_target_ts.is_missing,
            tf.fill(normalized_after_pre_intervention_period_length, True)
        ],
            axis=0))
    pre_intervention_target_ts_standard_deviation = tf.convert_to_tensor(
        np.nanstd(ci_data.pre_intervention_target_ts.time_series, ddof=1), dtype=dtype)
    # TODO(colcarroll,jburnim): Move constants to be module-level and commented.
    r2 = 0.8
    if design_matrix is not None:
        observation_noise_scale = (
                tf.cast(tf.math.sqrt(1 - r2), dtype=dtype) * pre_intervention_target_ts_standard_deviation)
    else:
        observation_noise_scale = pre_intervention_target_ts_standard_deviation
    level_scale = tf.ones([], dtype=dtype) * prior_level_sd * pre_intervention_target_ts_standard_deviation
    seasonal_drift_scales = 0.01 * pre_intervention_target_ts_standard_deviation * tf.ones(
        shape=[len(seasons)], dtype=dtype)
    if ci_data.normalized_whole_period_features is None:
        weights = tf.zeros([0], dtype=dtype)
    else:
        weights = tf.zeros(ci_data.normalized_whole_period_features.shape[-1:], dtype=dtype)

    level = tf.zeros_like(extended_target_ts.time_series)
    slope = tf.zeros_like(extended_target_ts.time_series)

    samples, posterior_means, posterior_predictive_samples = _run_gibbs_sampler(
        sts_model=model,
        outcome_ts=extended_target_ts,
        outcome_sd=pre_intervention_target_ts_standard_deviation,
        design_matrix=design_matrix,
        num_results=num_results,
        num_warmup_steps=num_warmup_steps,
        observation_noise_scale=observation_noise_scale,
        level_scale=level_scale,
        seasonal_drift_scales=seasonal_drift_scales,
        weights=weights,
        level=level,
        slope=slope,
        seed=seed,
        dtype=dtype,
        seasons=seasons,
        # By passing in an extra Python integer, it will be part of the cache
        # key. Calling this twice with the same
        # experimental_tf_function_cache_key_addition will result in a cache hit,
        # while changing it to a (previously unused) different value will result
        # in a miss.
        experimental_tf_function_cache_key_addition=experimental_tf_function_cache_key_addition
    )
    return samples, posterior_means, posterior_predictive_samples


def _get_posterior_means_and_trajectories(sts_model, posterior_samples, seed):
    """Computes means and draws from posterior.

    Args:
      sts_model: Model that was used to fit the samples.
      posterior_samples: Draws from posterior of sts_model.
      seed: PRNG seed.

    Returns:
      Posterior means, posterior predictive samples
    """
    predictive_distributions = gibbs_sampler.one_step_predictive(
        sts_model,
        posterior_samples,
        # This gets us a sample per posterior sample, rather than just a fraction
        # of them.
        thin_every=1,
        use_zero_step_prediction=True)
    posterior_means = predictive_distributions.mean()
    # Returns a shape of [num_timesteps, num_posterior_samples].
    output_sample = predictive_distributions.components_distribution.sample(
        seed=seed)
    posterior_trajectories = tf.transpose(output_sample)
    return posterior_means, posterior_trajectories


def _compute_impact(
        posterior_means,
        posterior_trajectories,
        ci_data: cid.CausalImpactData,
        alpha: float = 0.05,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Computes impact using posterior predictive distributions.

    Adds two attributes, both pd.DataFrames: one with timepoint-wise impact
    estimates, and another containing summaries over the post-treatment period.
    The former is used for plotting, the latter for summarizing overall impact.

    Args:
      posterior_means: Mean of posterior predictive trajectories.
      posterior_trajectories: Draws from the posterior predictive distribution.
        variables from sts_model.
      ci_data: an instance of a cid.CausalImpactData object.
      alpha: float, used to get confidence level for predictions.

    Returns:
      Two dataframes, the first with the impact timeseries, the second with a
      summary of that data.
    """
    # Check alpha.
    if not 0 < alpha < 1:
        raise ValueError("`alpha` must be between 0 and 1.")

    # Extract attributes of data object that we'll need for computing impact.
    observed_ts_pre = ci_data.pre_intervention_data[ci_data.target_col]
    observed_ts_post = ci_data.after_pre_intervention_data[ci_data.target_col]
    # Filter out data after the post-period.
    observed_ts_post = observed_ts_post.loc[
        (observed_ts_post.index >= ci_data.post_intervention_period[0])
        & (observed_ts_post.index <= ci_data.post_intervention_period[1])]
    observed_ts_full = pd.concat([observed_ts_pre, observed_ts_post], axis=0)

    # Get samples of posterior predictive trajectories and posterior means.
    # Ensure that lower quantile comes first.
    quantiles = (alpha / 2.0, 1.0 - (alpha / 2.0))
    posterior_trajectories, posterior_trajectory_summary = (
        _sample_posterior_predictive(
            posterior_means=posterior_means,
            posterior_trajectories=posterior_trajectories,
            ci_data=ci_data,
            quantiles=quantiles))

    # Use the posterior samples of trajectories to compute corresponding
    # trajectories for point, cumulative point, and relative effect estimates.
    trajectory_dict = _compute_impact_trajectories(
        posterior_trajectories,
        observed_ts_full,
        treatment_start=ci_data.post_intervention_period[0])

    # Create time series of mean and lower/upper quantiles for the point and
    # cumulative predictions.
    series = _compute_impact_estimates(
        posterior_trajectory_summary=posterior_trajectory_summary,
        trajectory_dict=trajectory_dict,
        observed_ts_full=observed_ts_full,
        ci_data=ci_data,
        quantiles=quantiles)

    # Create table to summarize results over the entire post-period.
    summary = _compute_summary(
        posterior_trajectory_summary=posterior_trajectory_summary,
        trajectory_dict=trajectory_dict,
        observed_ts_post=observed_ts_post,
        post_period=ci_data.post_intervention_period,
        quantiles=quantiles,
        alpha=alpha)
    return series, summary


def _sample_posterior_predictive(
        posterior_means: TensorLike,
        posterior_trajectories: TensorLike,
        ci_data: cid.CausalImpactData,
        quantiles: Tuple[float, float]
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Samples from posterior predictive distribution and calculates summaries.

    Draws sample trajectories from the posterior predictive distribution and
    calculates the posterior mean for each time point. The posterior means are
    calculated separately for each component of the structural time series, so
    they exclude observation noise. Note that the mean of the posterior
    trajectories is, in expectation, equal to the posterior mean, but any given
    simulation with a finite number of MCMC iterations will include observation
    noise, so here we calculate the expectation without observation noise.
    This function also calculates posterior quantiles for each time point
    using the sampled trajectories at the values given in `quantiles`.

    Args:
      posterior_means: Means of the posterior predictive trajectories.
      posterior_trajectories: Draws from the posterior predictive distribution.
        variables from sts_model.
      ci_data: an instance of a cid.CausalImpactData object.
      quantiles: floats in (0, 1) giving the quantiles to calculate.

    Returns:
      Tuple of:
        posterior_trajectories: Sampled posterior trajectories. Columns represent
          the sampled posterior trajectories.
        posterior_trajectory_summary: Summary of the trajectories, including
          mean and specified quantiles. Columns represent the summary statistics.
      In both dataframes, the rows represent time and have the same index as the
      original data.
    """
    # Check that given quantiles are in (0, 1) and in order.
    if any((q < 0) | (q > 1) for q in quantiles):
        raise ValueError("All elements of `quantiles` must be in (0, 1). Got %s" %
                         quantiles)
    if quantiles[0] > quantiles[1]:
        raise ValueError("`quantiles` must be sorted in ascending order. Got %s" %
                         quantiles)

    # Compute posterior mean trajectories in the pre- and post-periods from the
    # predictive distributions.
    posterior_means = posterior_processing.process_posterior_quantities(
        ci_data, posterior_means, ["posterior_mean"])

    # Sample posterior trajectories from the posterior predictive distributions
    # and calculate the specified quantiles at each time point.
    posterior_trajectories = _package_posterior_trajectories(
        posterior_trajectories, ci_data)

    posterior_quantiles = posterior_processing.calculate_trajectory_quantiles(
        posterior_trajectories, "posterior", quantiles)

    # Combine posterior means and quantiles into a single dataframe by joining
    # on the index.
    posterior_trajectory_summary = posterior_means.join(posterior_quantiles)

    return (posterior_trajectories, posterior_trajectory_summary)


def _package_posterior_trajectories(
        posterior_trajectories: TensorLike,
        ci_data: cid.CausalImpactData) -> pd.DataFrame:
    """Repackages trajectories into a dataframe.

    NOTE: if the data in ci_data.model data was scaled, this function will undo
    the scaling so that the posterior trajectories are on the *original* scale.

    Args:
      posterior_trajectories: Sampled posterior predictive trajectories.
      ci_data: an instance of a cid.CausalImpactData object.

    Returns:
      pd.DataFrame with rows corresponding to time points and columns
        corresponding to each sampled posterior trajectory.
    """
    col_names = [
        f"sample_{i + 1}" for i in range(posterior_trajectories.shape[0])
    ]
    return posterior_processing.process_posterior_quantities(
        ci_data, posterior_trajectories, col_names)


def _compute_impact_trajectories(
        posterior_trajectories: pd.DataFrame, observed_ts_full: pd.Series,
        treatment_start: OutputDateType) -> Dict[str, pd.DataFrame]:
    """Computes trajectories of point and cumulative effects.

    Uses the sampled posterior predictive trajectories and observed outcome time
    series to compute corresponding sample trajectories of:
    point effects (observed - predicted),
    cumulative point effects (cumulative sum(observed - predicted))

    Args:
      posterior_trajectories: pd.DataFrame with rows corresponding to time points
        and columns corresponding to each sampled posterior trajectory.
      observed_ts_full: the observed outcome for the full time period.
      treatment_start: start time of the treatment.

    Returns:
      dictionary of the posterior prediction, point, and cumulative effect
      trajectories.
    """

    # Point effects are zero in the pre-period by definition. However, we do not
    # enforce this restriction here so that the pre-period point effect
    # estimates can be used as a model check (the uncertainty intervals should
    # include zero) and as a check of the treatment start date.
    # To calculate point effects, we first subtract the observed data from the
    # trajectories and then multiply by -1 to get the correct quantity. This
    # way, we avoid having to replicate the observed data to match the shape of
    # the sampled trajectories.
    point_effect_trajectories = posterior_trajectories.sub(
        observed_ts_full, axis=0).mul(-1)

    # Cumulative point effects are zero in the pre-period by definition.
    cum_effect_trajectories_base = point_effect_trajectories.copy()
    cum_effect_trajectories_base.loc[
        cum_effect_trajectories_base.index < treatment_start] = 0
    # Use axis=0 to calculate cumulative sum over time (down the rows) for each
    # sample (column).
    cum_effect_trajectories = cum_effect_trajectories_base.cumsum(axis=0)

    return {
        "predictions": posterior_trajectories,
        "point_effects": point_effect_trajectories,
        "cumulative_effects": cum_effect_trajectories
    }


def _compute_impact_estimates(posterior_trajectory_summary: pd.DataFrame,
                              trajectory_dict: Dict[str, pd.DataFrame],
                              observed_ts_full: pd.Series,
                              ci_data: cid.CausalImpactData,
                              quantiles: Tuple[float, float]) -> pd.DataFrame:
    """Computes timepoint-wise summaries of predictions and impact estimates.

    Takes the posterior means and trajectories calculated in predict() and
    computes point and cumulative point effects at each time point by
    subtracting from the observed data. This function calculates means and
    quantiles for point and cumulative point effects.

    The output of this function is the input to creating impact plots.

    Args:
      posterior_trajectory_summary: pd.DataFrame output of predict() containing
        the posterior means and quantiles of the predictions.
      trajectory_dict: dictionary of the posterior prediction trajectories and
        point, cumulative, and relative effect trajectories as calculated in
        _compute_impact_trajectories().
      observed_ts_full: the observed outcome for the full time period.
      ci_data: an instance of a cid.CausalImpactData object.
      quantiles: floats in (0, 1) giving the quantiles to calculate.

    Returns:
      pd.DataFrame with the same index as the original data and
        columns for mean, lower quantile, and upper quantile for the point
        predictions, point effects (i.e. observed data minus point prediction),
        and cumulative effects (i.e. cumulative observed data minus cumulative
        point prediction). It will also have columns for the start and end
        points of the pre and post periods.
    """
    # Using the posterior means and observed time series, calculate mean point
    # and cumulative point effects. Note that cumulative effects are zero in the
    # pre-period by definition.
    point_effects_mean = (
            observed_ts_full - posterior_trajectory_summary["posterior_mean"])
    point_effects_mean = point_effects_mean.to_frame(name="point_effects_mean")
    cum_effects_mean_base = point_effects_mean.copy()
    zero_inds = point_effects_mean.index < ci_data.post_intervention_period[0]
    cum_effects_mean_base.loc[zero_inds] = 0
    cum_effects_mean = cum_effects_mean_base.cumsum()
    cum_effects_mean.columns = ["cumulative_effects_mean"]

    # Calculate quantiles at each time point for point and cumulative effect
    # trajectories.
    point_effects_quantiles = posterior_processing.calculate_trajectory_quantiles(
        trajectory_dict["point_effects"], "point_effects", quantiles)
    cum_effects_quantiles = posterior_processing.calculate_trajectory_quantiles(
        trajectory_dict["cumulative_effects"], "cumulative_effects", quantiles)

    # Collect all dataframes
    impact_estimates = pd.concat([
        observed_ts_full.to_frame(name="observed"), posterior_trajectory_summary,
        point_effects_mean, point_effects_quantiles, cum_effects_mean,
        cum_effects_quantiles
    ],
        axis=1)

    # The in-between period and after post-period should only have observed and
    # posteriors (to match original).
    impact_estimates.loc[
        ((impact_estimates.index > ci_data.pre_intervention_period[1]) &
         (impact_estimates.index < ci_data.post_intervention_period[0])) |
        (impact_estimates.index > ci_data.post_intervention_period[1]),
        impact_estimates.columns.difference(
            ["observed", "posterior_mean", "posterior_lower", "posterior_upper"]
        )] = np.nan

    # Where the observed was NaN, all other values should be NaN (and not zero).
    # This follows from the fact there is no real value to compare to.
    impact_estimates.loc[
        np.isnan(impact_estimates["observed"]),
        impact_estimates.columns.difference(
            ["observed", "posterior_mean", "posterior_lower", "posterior_upper"]
        )] = np.nan

    # The impact estimates so far of the range [pre_period[0], end_of_input_data],
    # but we actually want to return the original values before the pre-period.
    # Thus change the index to match the original data, then copy over the entire
    # original timeseries.
    impact_estimates = impact_estimates.reindex(
        ci_data.data.index, copy=False, fill_value=np.nan)
    impact_estimates["observed"] = ci_data.data[ci_data.target_col]

    # Add the pre/post period dates as columns for easier plotting.
    impact_estimates["pre_period_start"] = ci_data.pre_intervention_period[0]
    impact_estimates["pre_period_end"] = ci_data.pre_intervention_period[1]
    impact_estimates["post_period_start"] = ci_data.post_intervention_period[0]
    impact_estimates["post_period_end"] = ci_data.post_intervention_period[1]

    return impact_estimates


def _compute_summary(posterior_trajectory_summary: pd.DataFrame,
                     trajectory_dict: Dict[str, pd.DataFrame],
                     observed_ts_post: pd.Series, post_period: OutputPeriodType,
                     quantiles: Tuple[float,
                     float], alpha: float) -> pd.DataFrame:
    """Computes summary statistics of the average and cumulative impact.

    Calculates average and cumulative summaries of counterfactual forecasts and
    pointwise impact estimates over the post-treatment period. Note that because
    the average/sum of a quantile isn't the same as the quantile of an
    average/sum, we can't just use the output of _compute_impact_estimates()
    and instead need to use the sampled trajectories to get the correct values.

    Args:
      posterior_trajectory_summary: pd.DataFrame output of predict() containing
        the posterior means and quantiles of the predictions.
      trajectory_dict: dictionary of the posterior prediction trajectories and
        point and cumulative effect trajectories as calculated in
        _compute_impact_trajectories().
      observed_ts_post: the observed outcome for the post-period.
      post_period: Range of the post-period.
      quantiles: floats in (0, 1) giving the quantiles to calculate.
      alpha: float specifying the quantiles.

    Returns:
      pd.DataFrame with the same index as the original data and
        columns for mean, lower quantile, and upper quantile for the point
        predictions, point effects (i.e. observed data minus point prediction),
        and cumulative effects (i.e. cumulative observed data minus cumulative
        point prediction).
    """

    # Restrict the posterior mean trajectory and sampled impact trajectories to
    # the post-period.
    posterior_mean = posterior_trajectory_summary.loc[
        (posterior_trajectory_summary.index >= post_period[0]) &
        (posterior_trajectory_summary.index <= post_period[1]), "posterior_mean"]
    trajectory_dict = {
        k: v.loc[(v.index >= post_period[0]) & (v.index <= post_period[1])]
        for k, v in trajectory_dict.items()
    }

    # PREDICTIONS
    # Calculate average and cumulative values of the posterior mean -- i.e. the
    # counterfactual predictions -- over the post-period.
    average_prediction = posterior_mean.mean()
    cumulative_prediction = posterior_mean.sum()

    # Use the sampled posterior trajectories to get uncertainty intervals for
    # the above quantities. We'll calculate the mean (sum) of the trajectories
    # *across the post-period*. Since we want to summarize over time for each
    # trajectory, we use axis=0.
    pred_trajectories_mean = trajectory_dict["predictions"].mean(axis=0)
    pred_trajectories_sum = trajectory_dict["predictions"].sum(axis=0)
    avg_pred_lower, avg_pred_upper = pred_trajectories_mean.quantile(quantiles)
    cum_pred_lower, cum_pred_upper = pred_trajectories_sum.quantile(quantiles)

    # POINT EFFECTS
    # Calculate the average and cumulative point effects. Use the corresponding
    # sampled impact trajectories to calculate lower/upper uncertainty
    # intervals.
    average_point_effect = observed_ts_post.mean() - average_prediction
    pt_eff_trajectories_mean = trajectory_dict["point_effects"].mean(axis=0)
    avg_pt_eff_lower, avg_pt_eff_upper = (
        pt_eff_trajectories_mean.quantile(quantiles))

    cumulative_point_effect = observed_ts_post.sum() - cumulative_prediction
    pt_eff_trajectories_sum = trajectory_dict["point_effects"].sum(axis=0)
    cum_pt_eff_lower, cum_pt_eff_upper = (
        pt_eff_trajectories_sum.quantile(quantiles))

    # RELATIVE EFFECTS
    # Calculate the average and cumulative relative effects. Use the
    # corresponding sampled impact trajectories to calculate lower/upper
    # uncertainty intervals. Note that the average and cumulative relative
    # effects are mathematically identical.
    rel_eff_trajectories_mean = (
            observed_ts_post.sum() / pred_trajectories_sum - 1.)
    avg_rel_eff_lower, avg_rel_eff_upper = (
        rel_eff_trajectories_mean.quantile(quantiles))

    cumulative_relative_effect = rel_eff_trajectories_mean
    cum_rel_eff_lower, cum_rel_eff_upper = (
        cumulative_relative_effect.quantile(quantiles))

    # Calculate summary statistics of the average and cumulative impact over the
    # post-period.
    summary_dict = {
        "actual": {
            "average": observed_ts_post.mean(),
            "cumulative": observed_ts_post.sum()
        },
        "predicted": {
            "average": average_prediction,
            "cumulative": cumulative_prediction
        },
        "predicted_lower": {
            "average": avg_pred_lower,
            "cumulative": cum_pred_lower
        },
        "predicted_upper": {
            "average": avg_pred_upper,
            "cumulative": cum_pred_upper
        },
        "predicted_sd": {
            "average": pred_trajectories_mean.std(),
            "cumulative": pred_trajectories_sum.std()
        },
        "abs_effect": {
            "average": average_point_effect,
            "cumulative": cumulative_point_effect
        },
        "abs_effect_lower": {
            "average": avg_pt_eff_lower,
            "cumulative": cum_pt_eff_lower
        },
        "abs_effect_upper": {
            "average": avg_pt_eff_upper,
            "cumulative": cum_pt_eff_upper
        },
        "abs_effect_sd": {
            "average": pt_eff_trajectories_mean.std(),
            "cumulative": pt_eff_trajectories_sum.std()
        },
        "rel_effect": {
            "average": np.mean(cumulative_relative_effect),
            "cumulative": np.mean(cumulative_relative_effect)
        },
        "rel_effect_lower": {
            "average": avg_rel_eff_lower,
            "cumulative": cum_rel_eff_lower
        },
        "rel_effect_upper": {
            "average": avg_rel_eff_upper,
            "cumulative": cum_rel_eff_upper
        },
        "rel_effect_sd": {
            "average": cumulative_relative_effect.std(),
            "cumulative": cumulative_relative_effect.std()
        }
    }
    summary_df = pd.DataFrame(summary_dict)

    # Add a column for the one-sided "p-value" based on the cumulative effect.
    # Note that the p-value will be the same for both the "average" and
    # "cumulative" rows. We concatenate the observed cumulative outcome with the
    # sampled ones to ensure that the p-value will fall within (0, 1).
    observed_cumulative_outcome = observed_ts_post.sum()
    sampled_cumulative_outcomes = pd.concat(
        [pred_trajectories_sum,
         pd.Series(observed_cumulative_outcome)], axis=0)
    prop_obs_lessthan = (observed_cumulative_outcome <=
                         sampled_cumulative_outcomes).mean()
    prop_obs_greaterthan = (observed_cumulative_outcome >=
                            sampled_cumulative_outcomes).mean()
    p_value = min(prop_obs_lessthan, prop_obs_greaterthan)
    summary_df["p_value"] = p_value
    summary_df["alpha"] = alpha

    return summary_df



# Copyright 2020-2023 The TFP CausalImpact Authors
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for working with (results from) the posterior."""

from typing import List, Text, Tuple, Union

from causalimpact import data as cid
import numpy as np
import pandas as pd


def calculate_trajectory_quantiles(
    trajectories: pd.DataFrame,
    column_prefix: Text = "predicted",
    quantiles: Tuple[float, float] = (0.025, 0.975)
) -> Union[pd.DataFrame, pd.Series]:
  """Calculates timepoint-wise quantiles of trajectories.

  This function is used to calculate timepoint-wise quantiles for both the
  posterior predictions and the cumulative predictions.

  Args:
    trajectories: pd.DataFrame of samples to take quantiles of. `trajectories`
      should have a DatetimeIndex (so rows are time points) and columns for each
      sample. The quantiles are calculated across columns (i.e. across samples
      for each time point) using the 'axis=1' argument in quantile().
    column_prefix: string giving the prefix to use for the column names of the
      quantiles; e.g. if the trajectories being passed are the cumulative ones,
      use column_prefix = "cumulative" and the returned dataframe will have
      columns "cumulative_lower" and "cumulative_upper".
    quantiles: floats in (0, 1) giving the quantiles to calculate.

  Returns:
    quantiles_df: dataframe with columns for lower/upper quantiles of
      `trajectories` and a corresponding DatetimeIndex.
  """

  column_names = [column_prefix + "_" + suffix for suffix in ["lower", "upper"]]
  # Use axis=1 to summarize across samples for each time point.
  # With two quantiles, quantile() returns a pd.DataFrame of size
  # 2 x (number of time points), so use transpose() to put time back in
  # the rows.
  quantiles_df = trajectories.quantile(q=quantiles, axis=1).transpose()
  quantiles_df.columns = column_names
  quantiles_df.index = trajectories.index

  return quantiles_df


def process_posterior_quantities(ci_data: cid.CausalImpactData,
                                 vals_to_process: np.ndarray,
                                 col_names: List[Text]) -> pd.DataFrame:
  """Process posterior quantities by undoing any scaling and reshaping.

  This function assumes that the input np.ndarray `vals_to_process` has one or
  more rows corresponding to posterior samples and columns corresponding to
  time points. IMPORTANT: this function assumes that the time points correspond
  to the full time period, pre- and post-period combined. The function does the
  following:
  * undoes any scaling, if needed
  * reshapes so that rows correspond to time points and columns to samples.
  * reformats as a pd.DataFrame with a DatetimeIndex and appropriate column
    names.

  Args:
    ci_data: an instance of a cid.CausalImpactData object.
    vals_to_process: the input array.
    col_names: list of column names to use in the output.

  Returns:
    pd.DataFrame with rows corresponding to time points and columns
    named according to `col_names`.
  """
  # If the data used for modeling were scaled, first undo the scaling.
  if ci_data.standardize_data:
    vals_to_process = ci_data.outcome_scaler.inverse_transform(vals_to_process)
  # Transpose so that rows are again time points and columns are samples.
  vals_to_process = np.transpose(vals_to_process)

  # Format as pd.DataFrame and add the appropriate DatetimeIndex. Use the union
  # of the pre/post data in case the pre/post periods don't cover the full
  # data time period.
  index = ci_data.normalized_pre_intervention_data.index.union(
      ci_data.normalized_after_pre_intervention_data.index).sort_values()
  return pd.DataFrame(vals_to_process, columns=col_names, index=index)



# Copyright 2020-2023 The TFP CausalImpact Authors
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotting causalimpact results."""

from typing import Any, Union

import altair as alt
import numpy as np
import pandas as pd
import tensorflow_probability as tfp
import matplotlib as mpl


def _draw_plotly_plot(plot_df, **plot_params):
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    """Make interactive Plotly plot."""
    # Extract plot parameters with defaults
    chart_width = plot_params.get("chart_width", 800)
    chart_height = plot_params.get("chart_height", 600)
    axis_title_font_size = plot_params.get("axis_title_font_size", 14)
    tick_font_size = plot_params.get("tick_font_size", 12)
    legend_font_size = plot_params.get("legend_font_size", 12)
    title_font_size = plot_params.get("title_font_size", 16)

    # Create subplots: 3 rows, 1 column
    fig = make_subplots(
        rows=3, cols=1,
        shared_xaxes=True,
        vertical_spacing=0.05,
        subplot_titles=(
            "Original vs Predicted",
            "Pointwise Effects",
            "Cumulative Effects"
        )
    )

    # Helper function to add vertical lines
    def add_vertical_lines(fig, pre_start, pre_end, post_start, post_end, row):
        shapes = []
        # Add lines conditionally
        if plot_df["time"].min() < pre_start:
            shapes.append(dict(
                type="line",
                x0=pre_start, y0=0, x1=pre_start, y1=1,
                xref='x', yref=f'y{row}',
                line=dict(color="grey", dash="dash")
            ))
        if plot_df["time"].min() < pre_end and plot_df["time"].max() > pre_end and pre_end < post_start:
            shapes.append(dict(
                type="line",
                x0=pre_end, y0=0, x1=pre_end, y1=1,
                xref='x', yref=f'y{row}',
                line=dict(color="grey", dash="dash")
            ))
        shapes.append(dict(
            type="line",
            x0=post_start, y0=0, x1=post_start, y1=1,
            xref='x', yref=f'y{row}',
            line=dict(color="grey", dash="dash")
        ))
        if plot_df["time"].max() > post_end:
            shapes.append(dict(
                type="line",
                x0=post_end, y0=0, x1=post_end, y1=1,
                xref='x', yref=f'y{row}',
                line=dict(color="grey", dash="dash")
            ))
        for shape in shapes:
            fig.add_shape(shape)

    # Extract period boundaries
    pre_period_start = plot_df["pre_period_start"].iloc[0]
    pre_period_end = plot_df["pre_period_end"].iloc[0]
    post_period_start = plot_df["post_period_start"].iloc[0]
    post_period_end = plot_df["post_period_end"].iloc[0]

    # Add vertical lines to each subplot
    add_vertical_lines(fig, pre_period_start, pre_period_end, post_period_start, post_period_end, row=1)
    add_vertical_lines(fig, pre_period_start, pre_period_end, post_period_start, post_period_end, row=2)
    add_vertical_lines(fig, pre_period_start, pre_period_end, post_period_start, post_period_end, row=3)

    # Plot Original and Predicted in the first subplot
    original_series = plot_df[(plot_df["scale"] == "original") & (plot_df["stat"] == "observed")]
    predicted_series = plot_df[(plot_df["scale"] == "original") & (plot_df["stat"] == "mean")]

    fig.add_trace(
        go.Scatter(
            x=predicted_series["time"],
            y=predicted_series["value"],
            mode='lines',
            name='Mean',
            line=dict(color='blue')
        ),
        row=1, col=1
    )

    fig.add_trace(
        go.Scatter(
            x=original_series["time"],
            y=original_series["value"],
            mode='lines',
            name='Observed',
            line=dict(color='orange')
        ),
        row=1, col=1
    )

    # Add confidence interval as a shaded area
    fig.add_trace(
        go.Scatter(
            x=predicted_series["time"],
            y=predicted_series["upper"],
            mode='lines',
            line=dict(width=0),
            showlegend=False
        ),
        row=1, col=1
    )

    fig.add_trace(
        go.Scatter(
            x=predicted_series["time"],
            y=predicted_series["lower"],
            mode='lines',
            fill='tonexty',
            fillcolor='rgba(0, 0, 255, 0.2)',
            line=dict(width=0),
            showlegend=False
        ),
        row=1, col=1
    )

    # Plot Pointwise Effects in the second subplot
    pointwise_series = plot_df[(plot_df["scale"] == "point_effects") & (plot_df["stat"] == "mean")]

    fig.add_trace(
        go.Scatter(
            x=pointwise_series["time"],
            y=pointwise_series["value"],
            mode='lines',
            name='Pointwise',
            line=dict(color='green')
        ),
        row=2, col=1
    )

    fig.add_trace(
        go.Scatter(
            x=pointwise_series["time"],
            y=pointwise_series["upper"],
            mode='lines',
            line=dict(width=0),
            showlegend=False
        ),
        row=2, col=1
    )

    fig.add_trace(
        go.Scatter(
            x=pointwise_series["time"],
            y=pointwise_series["lower"],
            mode='lines',
            fill='tonexty',
            fillcolor='rgba(0, 128, 0, 0.2)',
            line=dict(width=0),
            showlegend=False
        ),
        row=2, col=1
    )

    # Add horizontal line at y=0
    fig.add_trace(
        go.Scatter(
            x=[plot_df["time"].min(), plot_df["time"].max()],
            y=[0, 0],
            mode='lines',
            line=dict(color='grey', dash='dash'),
            showlegend=False
        ),
        row=2, col=1
    )

    # Plot Cumulative Effects in the third subplot
    cumulative_series = plot_df[(plot_df["scale"] == "cumulative_effects") & (plot_df["stat"] == "mean")]

    fig.add_trace(
        go.Scatter(
            x=cumulative_series["time"],
            y=cumulative_series["value"],
            mode='lines',
            name='Cumulative',
            line=dict(color='red')
        ),
        row=3, col=1
    )

    fig.add_trace(
        go.Scatter(
            x=cumulative_series["time"],
            y=cumulative_series["upper"],
            mode='lines',
            line=dict(width=0),
            showlegend=False
        ),
        row=3, col=1
    )

    fig.add_trace(
        go.Scatter(
            x=cumulative_series["time"],
            y=cumulative_series["lower"],
            mode='lines',
            fill='tonexty',
            fillcolor='rgba(255, 0, 0, 0.2)',
            line=dict(width=0),
            showlegend=False
        ),
        row=3, col=1
    )

    # Add horizontal line at y=0
    fig.add_trace(
        go.Scatter(
            x=[plot_df["time"].min(), plot_df["time"].max()],
            y=[0, 0],
            mode='lines',
            line=dict(color='grey', dash='dash'),
            showlegend=False
        ),
        row=3, col=1
    )

    # Update layout for better aesthetics
    fig.update_layout(
        width=chart_width,
        height=chart_height,
        showlegend=True,
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=1.02,
            xanchor="right",
            x=1
        ),
        font=dict(
            size=tick_font_size,
            family="Arial"
        ),
        title_text="Interactive Plot",
        hovermode='x unified'
    )

    # Update each subplot's axes
    for i in range(1, 4):
        fig.update_xaxes(
            title_text="Date",
            title_font=dict(size=axis_title_font_size, family="Arial"),
            tickfont=dict(size=tick_font_size),
            row=i, col=1
        )
        fig.update_yaxes(
            title_font=dict(size=axis_title_font_size, family="Arial"),
            tickfont=dict(size=tick_font_size),
            row=i, col=1
        )

    # Adjust subplot titles font size
    for annotation in fig['layout']['annotations']:
        annotation['font'] = dict(size=title_font_size, family="Arial")

    return fig


def _draw_matplotlib_plot(plot_df, **plot_params):
    """Make actual matplotlib plot."""
    try:
        import matplotlib as mpl
        import matplotlib.pyplot as mplt  # pylint: disable=g-import-not-at-top
        import japanize_matplotlib
        mpl.rcParams['font.family'] = 'Noto Sans JP'
    except ImportError as e:
        raise ImportError(
            "matplotlib is required for using it as plotting backend. Please"
            " install it first."
        ) from e

    def _helper_vertical_lines(plot_df, ax):
        # Vertical line at pre-period end and post-period end
        pre_period_start = plot_df["pre_period_start"][0]
        pre_period_end = plot_df["pre_period_end"][0]
        post_period_start = plot_df["post_period_start"][0]
        post_period_end = plot_df["post_period_end"][0]
        # Only draw a line at the start of the pre-period if there are points before
        # it.
        if any(plot_df["time"] < pre_period_start):
            ax.axvline(pre_period_start, color="grey", linestyle="--")
        # Only draw a line at the end of the pre-period if there are points between
        # it and the start of the post-period.
        if any(
                (plot_df["time"] > pre_period_end)
                & (plot_df["time"] < post_period_start)
        ):
            ax.axvline(pre_period_end, color="grey", linestyle="--")

        # Always draw when the post-period starts.
        ax.axvline(post_period_start, color="grey", linestyle="--")

        # Only draw line at the end of the post-period if there are points after it.
        if any(plot_df["time"] > post_period_end):
            ax.axvline(post_period_end, color="grey", linestyle="--")
        return ax

    # Calculate figure size in inches (assuming plot_params are in pixels)
    fig_width = plot_params["chart_width"] / 100
    fig_height = 3 * plot_params["chart_height"] / 100  # For 3 subplots

    # Setup base plot with constrained_layout for better spacing
    fig, axes = mplt.subplots(
        3,
        1,
        figsize=(fig_width, fig_height),
        sharex=True,
        constrained_layout=True,  # Use constrained_layout instead of tight_layout
    )

    # Enable grid for all subplots
    for ax in axes:
        ax.grid()

    # Apply any helper functions if necessary
    axes[0] = _helper_vertical_lines(plot_df, axes[0])
    axes[1] = _helper_vertical_lines(plot_df, axes[1])
    axes[2] = _helper_vertical_lines(plot_df, axes[2])

    # Set common x-label for the bottom subplot
    axes[2].set_xlabel(
        "Time",
        fontsize=plot_params.get("axis_title_font_size", 6),
        fontweight="bold"
    )

    # Plot the original data with pre-period and post-period marked
    original_series = plot_df.loc[
        (plot_df["scale"] == "original") & (plot_df["stat"] == "observed")
        ]
    predicted_series = plot_df.loc[
        (plot_df["scale"] == "original") & (plot_df["stat"] == "mean")
        ]
    axes[0].plot(
        predicted_series["time"], predicted_series["value"], label="Mean"
    )
    axes[0].plot(
        original_series["time"], original_series["value"], label="Observed"
    )
    axes[0].fill_between(
        original_series["time"],
        original_series["lower"],
        original_series["upper"],
        alpha=0.2,
    )
    # Place legend outside the plot area to the right
    axes[0].legend(
        loc="upper left",
        bbox_to_anchor=(1, 1),
        fontsize="small",
        frameon=False
    )

    # Plot the pointwise effect with pre-period and post-period marked
    pointwise_series = plot_df.loc[
        (plot_df["scale"] == "point_effects") & (plot_df["stat"] == "mean")
        ]
    axes[1].plot(
        pointwise_series["time"], pointwise_series["value"], label="Pointwise"
    )
    axes[1].fill_between(
        pointwise_series["time"],
        pointwise_series["lower"],
        pointwise_series["upper"],
        alpha=0.2,
    )
    # Add horizontal line at y=0
    axes[1].axhline(0, color="grey", linestyle="--")
    # Place legend outside the plot area to the right
    axes[1].legend(
        loc="upper left",
        bbox_to_anchor=(1, 1),
        fontsize="small",
        frameon=False
    )

    # Plot the cumulative effect with pre-period and post-period marked
    cumulative_series = plot_df.loc[
        (plot_df["scale"] == "cumulative_effects") & (plot_df["stat"] == "mean")
        ]
    axes[2].plot(
        cumulative_series["time"], cumulative_series["value"], label="Cumulative"
    )
    axes[2].fill_between(
        cumulative_series["time"],
        cumulative_series["lower"],
        cumulative_series["upper"],
        alpha=0.2,
    )
    # Add horizontal line at y=0
    axes[2].axhline(0, color="grey", linestyle="--")
    # Place legend outside the plot area to the right
    axes[2].legend(
        loc="upper left",
        bbox_to_anchor=(1, 1),
        fontsize="small",
        frameon=False
    )

    # Adjust layout to make space for legends
    # If using constrained_layout, this might not be necessary, but added for safety
    fig.subplots_adjust(right=0.85)

    return fig


def plot(ci_model, **kwargs) -> Union[alt.Chart, Any]:
    """Main plotting function.

    Args:
      ci_model: CausalImpactAnalysis object, after having called
        `fit_causalimpact`.
      **kwargs: arguments for modifying plot defaults:
        static_plot - whether to return the standard CausalImpact plot as a
          static plot (default) or an interactive plot.
        backend - literal["altair","matplotlib"] to use for generating the figure
        alpha - float for determining confidence level for uncertainty intervals
          when quantile_based_intervals=False.
        show_median - whether to draw posterior median predictions in addition to
          the posterior mean. Only applies if "median" was a specified aggregation
          given in evaluate().
        use_std_intervals - whether to draw uncertainty intervals based on
          quantiles (default) or use a normal approximation based on the standard
          deviation.
        chart_width - integer for chart width in pixels.
        chart_height - integer for chart height in pixels.
        axis_title_font_size - integer for axis title font size. Default = 18.
        axis_label_font_size - integer for axis title font size. Default = 16.
        strip_title_font_size - integer for facet label font size. Default = 18.

    Returns:
      alt.Chart plot object
    """

    # Process kwargs.
    plot_params = {
        "static_plot": True,
        "backend": "altair",
        "alpha": 0.05,
        "show_median": False,
        "use_std_intervals": False,
        "chart_width": 600,
        "chart_height": 200,
        "axis_title_font_size": 18,
        "axis_label_font_size": 16,
        "strip_title_font_size": 20
    }
    if kwargs:
        for k, v in plot_params.items():
            plot_params[k] = kwargs.get(k, v)

    # Create the dataframe that will be used to create the plot.
    main_plot_df = _create_plot_df(ci_model.series, plot_params["alpha"])

    # If use_std_intervals=True, use std to draw the uncertainty intervals,
    # otherwise use the quantile-based intervals. We drop the unnecessary
    # observations instead of keeping the requested ones because dates outside of
    # the pre/post period will have NaN values for band_method.
    if plot_params["use_std_intervals"]:
        plot_df = main_plot_df.loc[main_plot_df["band_method"] != "quantile"]
    else:
        plot_df = main_plot_df.loc[main_plot_df["band_method"] != "std"]

    # Include median if requested.
    if plot_params["show_median"]:
        plot_df = plot_df.loc[plot_df["stat"] != "median"]
        plot_df["stat_pretty"] = pd.Categorical(
            plot_df["stat_pretty"], categories=["Observed", "Mean"], ordered=True
        )

    # Create the requested plot type.
    if plot_params["backend"] == "altair":
        if plot_params["static_plot"]:
            plot_df = plot_df.loc[(plot_df["stat"] != "median")]
            plt = _draw_classic_plot(plot_df, **plot_params)
        else:
            plt = _draw_interactive_plot(plot_df, **plot_params)
    elif plot_params["backend"] == "matplotlib":
        plt = _draw_matplotlib_plot(plot_df, **plot_params)
    elif plot_params["backend"] == "plotly":
        plt = _draw_plotly_plot(plot_df, **plot_params)
    else:
        raise ValueError(
            "backend must be one of 'altair' or 'matplotlib'. Got"
            f" {plot_params['backend']}."
        )
    return plt


def _create_plot_df(series: pd.DataFrame, alpha: float = 0.05) -> pd.DataFrame:
    """Creates dataframe for use in plotting impact inferences.

    This function creates the data to be used in plotting the impact inferences.
    The plot layers lines for the observed data and predicted values along with
    uncertainty bands and then facets these depending on the scale (original,
    pointwise effect, or cumulative effect). The dataframe therefore has columns
    "time", "value", "lower", "upper", "scale", "stat" (observed/mean/median,
    if "median" was a requested aggregation). If "std" was a requested
    aggregation, the dataframe also contains "band_method" indicating whether the
    interval formed by "lower" and "upper" is calculated based on quantiles or
    a normal approximation using the predicted value +/- z_alpha/2 * std. The
    plot also includes horizontal zero lines for the pointwise and cumulative
    effect panels, so the dataframe has columns for these values too.

    Args:
      series: pd.DataFrame as output by sts_mod.evaluate(). Should be a time
        series containing observed and predicted outcome, as well as absolute and
        cumulative impact estimates.
      alpha: float determining the confidence level for std-based uncertainty
        intervals.

    Returns:
      pd.DataFrame indexed by time for creating plots. The `value` column is used
        to draw lines, while `upper` and `lower` denote the corresponding
        uncertainty bounds. The `scale` column denotes whether `value` is on the
        original, pointwise (absolute effect), or cumulative scale. The `stat`
        column denotes whether the quantity in `value` refers to the observed
        outcome or the posterior mean (or posterior median if "median" is
        in "aggregations"]). Both `scale` and `stat` have corresponding columns
        `scale_prett` and `stat_pretty` for making nice plot labels. There is also
        a `zero` column to draw a horizontal reference line for the absolute and
        cumulative effect plots.
    """
    series["time"] = series.index

    # Create dataframes for each component of the plot (lines and uncertainty
    # bands, and standard devation-based uncertainty bands if requested).
    lines_df = _create_plot_component_df(series, "lines")
    bands_df = _create_plot_component_df(series, "bands")
    if any(["std" in col for col in series.columns]):
        std_df = _create_plot_component_df(series, "std", alpha)
        bands_df = pd.concat([bands_df, std_df], axis=0, sort=True)

    # Merge the component dataframes. Use left join because lines_df will contain
    # observed data outside of the pre/post period intervals, whereas bands_df
    # will not.
    plot_df = lines_df.merge(
        bands_df,
        on=[
            "time", "scale", "pre_period_start", "pre_period_end",
            "post_period_start", "post_period_end"
        ],
        how="left")

    # Add a zero column so we can plot a zero line for the absolute and
    # cumulative scales, but set it to np.nan for the original scale so it
    # doesn't get plotted.
    plot_df["zero"] = 0
    plot_df.loc[plot_df["scale"] == "original", "zero"] = np.nan

    # Make nicer versions of the scale and stat variables for use as plot labels.
    plot_df["scale_pretty"] = [
        "Original" if m == "original" else
        "Pointwise" if m == "point_effects" else "Cumulative"
        for m in plot_df["scale"]
    ]
    plot_df["scale_pretty"] = pd.Categorical(
        plot_df["scale_pretty"],
        categories=["Original", "Pointwise", "Cumulative"],
        ordered=True)
    plot_df["stat_pretty"] = plot_df["stat"].str.capitalize()
    plot_df["stat_pretty"] = pd.Categorical(
        plot_df["stat_pretty"],
        categories=["Observed", "Mean", "Median"],
        ordered=True)

    return plot_df


def _create_plot_component_df(series: pd.DataFrame,
                              component: str,
                              alpha: float = 0.05) -> pd.DataFrame:
    """Creates plot component dataframes.

    This function takes the impact estimate time series and creates a long-form
    dataframe that can be used to plot lines or uncertainty bands for estimates
    on the original data scale, pointwise (absolute) effects, and cumulative
    effects. The dataframe is indexed by time and has columns denoting the value
    to be plotted on the y-axis, the scale that that value is on (original,
    absolute, cumulative), and the statistic that the value corresponds to (e.g.
    the upper or lower bound, the observed data, the posterior mean or median).

    Args:
      series: dataframe containing impact estimates.
      component: string for which component dataframe ("lines", "bands", "std") to
        create. Note: "bands" denotes the standard quantile-based uncertainty
        bands, while "std" denotes standard deviation-based bands.
      alpha: float determining the confidence level for std-based uncertainty
        intervals.

    Returns:
      pd.DataFrame indexed by time, with the columns `value`, `scale` (whether
        `value` is on the original, pointwise (absolute effect), or cumulative
        scale). If `component` is "lines", then the dataframe also has columns for
        `stat`, the quantity to be plotted a a line (observed, posterior mean, or
        posterior median [only if "median" is in "aggregations"]). If `component`
        is "bands" or "std", then there are columns `upper` and `lower` denoting
        the bounds of the uncertainty intervals. In this case, there is also
        a column indicating which of "bands" or "std" was used for creating the
        upper/lower bounds.
    """

    if all([x not in component for x in ["lines", "bands", "std"]]):
        raise ValueError("`component` must be one of 'lines', 'bands', or 'std'."
                         "Got %s." % component)

    # Pull out only the columns we need from `series`.
    col_stubs = [
        "time", "pre_period_start", "pre_period_end", "post_period_start",
        "post_period_end"
    ]
    if component == "lines":
        col_stubs.extend(["mean", "median", "observed"])
    elif component == "bands":
        col_stubs.extend(["lower", "upper"])
    else:  # component == "std"
        col_stubs.extend(["mean", "std"])
    cols_to_extract = [
        col for col in series.columns if any(x in col for x in col_stubs)
    ]

    # Keep the necessary columns and reshape so that `scale` and `stat` are long.
    # To do this, we collapse all of the existing columns into a single column
    # that we then split into two columns, `scale` and `stat`.
    sub_df = series[cols_to_extract].melt(
        id_vars=[
            "time", "pre_period_start", "pre_period_end", "post_period_start",
            "post_period_end"
        ],
        var_name="scale_stat",
        value_name="value")

    # `scale` is for whether `value` is on the original scale (scale of the
    # observed data), the pointwise scale, or the cumulative scale.
    stats_to_drop = "_upper|_lower|_mean|_median|_std"
    sub_df["scale"] = sub_df["scale_stat"].str.replace(
        stats_to_drop, "", regex=True)
    sub_df.loc[sub_df["scale"].str.contains("observed|posterior"),
    "scale"] = "original"

    # `stat` is for whether `value` represents the observed data or the mean or
    # median estimates.
    sub_df["stat"] = sub_df["scale_stat"].str.replace(
        "posterior_|point_effects_|cumulative_effects_", "", regex=True)
    sub_df.drop(columns=["scale_stat"], inplace=True)

    # For bands and std, reshape the data again so that "upper" and "lower" (for
    # bands) and "mean" and "std" (for std) are wide.
    # columns.
    if (component == "bands") | (component == "std"):
        sub_df = sub_df.pivot_table(
            index=[
                "time", "scale", "pre_period_start", "pre_period_end",
                "post_period_start", "post_period_end"
            ],
            columns="stat",
            values="value").reset_index()
        if component == "bands":
            sub_df["band_method"] = "quantile"
        else:
            sub_df["band_method"] = "std"

    # For std, use the posterior mean and std to create lower and upper bounds.
    if component == "std":
        z_val = tfp.distributions.Normal(
            loc=0., scale=1.).quantile(1.0 - alpha / 2.0).numpy()
        sub_df["lower"] = sub_df["mean"] - z_val * sub_df["std"]
        sub_df["upper"] = sub_df["mean"] + z_val * sub_df["std"]
        sub_df.drop(columns={"mean", "std"}, inplace=True)

    return sub_df


def _create_base_layers(plot_df: pd.DataFrame, **kwargs):
    """Create base plot layers.

    This helper function is used by both _draw_static_plot() and
    _draw_interactive_plot() to draw the four base plot layers: lines, uncertainty
    bands, and horizontal/vertical rules for zero and the treatment start,
    respectively.

    Args:
      plot_df: dataframe to use for plotting.
      **kwargs: Optional plot parameters. chart_width - integer for chart width in
        pixels. Default = 600. chart_height - integer for chart height in pixels.
        Default = 200. axis_title_font_size - integer for axis title font size.
        Default = 18. axis_label_font_size - integer for axis title font size.
        Default = 16. strip_title_font_size - integer for facet label font size.
        Default = 20.

    Returns:
      Dictionary containing the four base plot layers: lines, bands, and vertical
        and horizontal lines. NOTE that the lines, bands, and horizontal line are
        alt.Chart objects, while the vertical lines, keyed by "vlines", are
        a subdictionary of alt.Chart objects. This subdictionary is keyed by the
        date at which the vertical line is to be drawn and the values are the
        corresponding alt.Chart objects for those vertical lines.
    """

    # Base line layer.
    base_lines = alt.Chart(plot_df).mark_line().encode(
        x=alt.X("time", title="Time"),
        y=alt.Y("value:Q", scale=alt.Scale(zero=False), title="")).properties(
        width=kwargs["chart_width"], height=kwargs["chart_height"])

    # Base band layer.
    base_band = alt.Chart(plot_df).mark_area(opacity=0.3).encode(
        x=alt.X("time", title="Time"), y="upper:Q", y2="lower:Q").properties(
        width=kwargs["chart_width"], height=kwargs["chart_height"])

    # Add horizontal line at zero.
    base_hline = alt.Chart(plot_df).mark_rule().encode(y="zero")

    # Add vertical lines at the edges of the pre-period and post-period whenever
    # there is something to demarcate.
    base_vlines = {}
    # Since pre_period_end and post_period_start are columns in plot_df, we can
    # just take the values in the first row in those columns.
    pre_period_start = plot_df["pre_period_start"][0]
    pre_period_end = plot_df["pre_period_end"][0]
    post_period_start = plot_df["post_period_start"][0]
    post_period_end = plot_df["post_period_end"][0]

    # Only draw a line at the start of the pre-period if there are points before
    # it.
    if any(plot_df["time"] < pre_period_start):
        base_vlines["pre_period_start"] = alt.Chart(plot_df).mark_rule(
            strokeDash=[5, 5]).encode(
            x=alt.X("pre_period_start"), color=alt.value("grey"))
    # Only draw a line at the end of the pre-period if there are points between
    # it and the start of the post-period.
    if any((plot_df["time"] > pre_period_end)
           & (plot_df["time"] < post_period_start)):
        base_vlines["pre_period_end"] = alt.Chart(plot_df).mark_rule(
            strokeDash=[5, 5]).encode(
            x=alt.X("pre_period_end"), color=alt.value("grey"))

    # Always draw when the post-period starts.
    base_vlines["post_period_start"] = alt.Chart(plot_df).mark_rule(
        strokeDash=[5, 5]).encode(
        x=alt.X("post_period_start"), color=alt.value("grey"))

    # Only draw line at the end of the post-period if there are points after it.
    if any(plot_df["time"] > post_period_end):
        base_vlines["post_period_end"] = alt.Chart(plot_df).mark_rule(
            strokeDash=[5, 5]).encode(
            x=alt.X("post_period_end"), color=alt.value("grey"))

    # Return the base plot components.
    return {
        "lines": base_lines,
        "band": base_band,
        "hline": base_hline,
        "vlines": base_vlines
    }


def _draw_classic_plot(plot_df: pd.DataFrame, **kwargs) -> alt.Chart:
    """Draw the classic static impact plot as in the R package.

    Args:
      plot_df: dataframe to use for plotting.
      **kwargs: Optional plot parameters. chart_width - integer for chart width in
        pixels. Default = 600. chart_height - integer for chart height in pixels.
        Default = 200. axis_title_font_size - integer for axis title font size.
        Default = 18. axis_label_font_size - integer for axis title font size.
        Default = 16. strip_title_font_size - integer for facet label font size.
        Default = 20.

    Returns:
      alt.Chart object.
    """
    layers = _create_base_layers(plot_df, **kwargs)

    # Add color spec to lines.
    color_spec = alt.Color(
        "stat_pretty:N",
        legend=alt.Legend(
            title="",
            labelFontSize=kwargs["axis_label_font_size"],
            symbolSize=10 * kwargs["axis_label_font_size"]))
    layers["lines"] = layers["lines"].encode(color=color_spec)

    # Unpack the vertical rule chart objects into a list; combine with the other
    # chart layers into a tuple that can be passed to alt.layer() to create the
    # final plot.
    vlines = list(layers["vlines"].values())
    chart_layers = tuple([layers["lines"], layers["band"], layers["hline"]] +
                         vlines)
    final_plot = alt.layer(
        *chart_layers, data=plot_df).facet(
        row=alt.Row(
            "scale_pretty:N",
            sort=["Original", "Pointwise", "Cumulative"],
            title="")).resolve_scale(y="independent").configure(
        background="white").configure_axis(
        titleFontSize=kwargs["axis_title_font_size"],
        labelFontSize=kwargs["axis_label_font_size"]
    ).configure_header(
        labelFontSize=kwargs["strip_title_font_size"])
    return final_plot


def _draw_interactive_plot(plot_df: pd.DataFrame, **kwargs) -> alt.Chart:
    """Draw interactive impact plot.

    Args:
      plot_df: dataframe to use for plotting
      **kwargs: Optional plot parameters. chart_width - integer for chart width in
        inches. Default = 15. chart_height - integer for chart height. Default =
        15. axis_title_font_size - integer for axis title font size. Default = 18.
        axis_label_font_size - integer for axis title font size. Default = 16.
        strip_title_font_size - integer for facet label font size. Default = 20.

    Returns:
      altair Chart object.
    """

    # ############################################################################
    # Create interactive selection elements.
    # ############################################################################

    # Brush for selecting a location to zoom into on x-axis.
    brush = alt.selection(type="interval", encodings=["x"])

    # Mini-chart as interactive legend to choose which stat to display.
    stat_selection = alt.selection_multi(fields=["stat_pretty"])
    selection_color = alt.condition(stat_selection,
                                    alt.Color("stat_pretty:N", legend=None),
                                    alt.value("lightgray"))
    legend = alt.Chart(plot_df).mark_point().encode(
        y=alt.Y("stat_pretty:N", axis=alt.Axis(orient="right"), title=""),
        color=selection_color).add_selection(stat_selection)

    # ############################################################################
    # Create the static top chart.
    # ############################################################################

    # The static chart is for the data on the original scale.
    static_df = plot_df.loc[plot_df["scale"] == "original"]

    # Create static layers and add the color spec to the lines layer and the
    # brush selection to the bands layer.
    static_layers = _create_base_layers(static_df, **kwargs)
    static_color_spec = alt.Color(
        "stat_pretty:N",
        legend=alt.Legend(
            title="",
            labelFontSize=kwargs["axis_label_font_size"],
            symbolSize=10 * kwargs["axis_label_font_size"]))
    static_layers["lines"] = static_layers["lines"].encode(
        color=static_color_spec)
    static_layers["band"] = static_layers["band"].add_selection(brush)

    # Combine layers of static top chart. Add the brush selection to the band
    # layer to enable selecting a date range on the x-axis for the bottom plots to
    # zoom in on (you only need to put it on one layer, so it could also have been
    # added to the lines layer).
    row_spec = alt.Row(
        "scale_pretty:N", sort=["Original", "Pointwise", "Cumulative"], title="")
    # Unpack the vertical rule chart objects into a list; combine with the other
    # chart layers into a tuple that can be passed to alt.layer to create the
    # full static plot.
    static_vlines = list(static_layers["vlines"].values())
    top_chart_layers = tuple(
        [static_layers["lines"], static_layers["band"], static_layers["hline"]] +
        static_vlines)
    top_static_plot = alt.layer(
        *top_chart_layers,
        data=static_df).facet(row=row_spec).resolve_scale(y="independent")

    # ############################################################################
    # Create the dynamic charts that will zoom upon selection on the static top
    # chart.
    # ############################################################################

    # Create dynamic layers and add interactive selections.
    dynamic_layers = _create_base_layers(plot_df, **kwargs)
    dynamic_layers["lines"] = dynamic_layers["lines"].encode(
        color=selection_color,
        x=alt.X("time", scale=alt.Scale(domain=brush), title="Time"))
    dynamic_layers["band"] = dynamic_layers["band"].encode(
        x=alt.X("time", scale=alt.Scale(domain=brush), title="Time"))

    # Add interactive selections to each of the vertical line chart objects.
    for vline_date, vline_object in dynamic_layers["vlines"].items():
        dynamic_layers["vlines"][vline_date] = vline_object.encode(
            x=alt.X(vline_date, scale=alt.Scale(domain=brush)))

    # Combine the dynamic chart layers into a tuple that can be used with
    # alt.layer() to create the full dynamic plot.
    dynamic_vlines = list(dynamic_layers["vlines"].values())
    bottom_chart_layers = tuple([
                                    dynamic_layers["lines"], dynamic_layers["band"], dynamic_layers["hline"]
                                ] + dynamic_vlines)
    bottom_dynamic_plot = alt.layer(
        *bottom_chart_layers,
        data=plot_df).facet(row=row_spec).resolve_scale(y="independent")

    # ############################################################################
    # Create the final chart.
    # ############################################################################

    # First, vertically concatenate the static and dynamic charts, then
    # horizontally concatenate with little interactive legend chart.
    final_chart = alt.vconcat(top_static_plot, bottom_dynamic_plot)
    return (final_chart | legend).configure(background="white").configure_axis(
        titleFontSize=kwargs["axis_title_font_size"],
        labelFontSize=kwargs["axis_label_font_size"]).configure_header(
        labelFontSize=kwargs["strip_title_font_size"])
